
#include "ServiceDiscoveryIfce.h"
#include "Common.h"
#include "Definitions.h"


void sd_bdii_freeService(SDService *service);
void sd_bdii_freeServiceList(SDServiceList *list);
void sd_bdii_freeServiceDataList(SDServiceDataList *list);
void sd_bdii_freeServiceDetails(SDServiceDetails *details);
void sd_bdii_freeServiceDetailsList(SDServiceDetailsList *servicedetailsList);

int get_details(const char *q, const char *name, LDAP *ld, LDAPMessage **reply,
        SDException *exception);
int parse_detail_attrs(SDServiceDetails *details, LDAP *ld, LDAPMessage *msg,
        SDException *exception);	
int add_to_list_bdii(SDServiceList *list, SDService *service, SDException *ex);	
int add_vo(SDServiceDetails *details, const char *vo);
int match_vos(LDAP *ld, LDAPMessage *msg, const SDVOList *vos);
int match_service_bdii(SDServiceDetails *details, const char *type,
        const char *site, const SDVOList *vos);
int match_data(const SDServiceDataList *list,
        const SDServiceDataList *requirements);
void add_sa_root(LDAP *ld, SDServiceDataList *list, const char *endpoint,
        SDException *exception);	
SDServiceDataList *query_service_data(LDAP *ld, LDAPMessage *imsg,
        const char *type, const char *endpoint, SDException *exception);
int parse_service_attrs(SDService *service, LDAP *ld, LDAPMessage *msg,
        SDException *exception);
	
extern char * default_search_base; 	
	
/**********************************************************************
 * Memory management methods
 */

void sd_bdii_freeService(SDService *service) {
    if (!service)
        return;

    if(service->name){
	    g_free(service->name);
	    service->name = NULL;
	    }
    if(service->type){
	    g_free(service->type);
	    service->type = NULL;
	    }
    if(service->endpoint)    {
	    g_free(service->endpoint);
	    service->endpoint = NULL;
	    }
    if(service->version) {   	    
	    g_free(service->version);
	    service->version = NULL;
	    }

    g_free(service);
    service = NULL;
}

void sd_bdii_freeServiceList(SDServiceList *list) {
    int i;

    if (!list)
        return;

    for (i = 0; i < list->numServices; i++)
        sd_bdii_freeService(list->services[i]);
    g_free(list->services);
    g_free(list);
}

void sd_bdii_freeServiceDataList(SDServiceDataList *list) {
    int i;

    if (!list)
        return;
	
    for (i = 0; i < list->numItems; i++) {
        g_free(list->items[i].key);
        g_free(list->items[i].value);
    }
    g_free(list->items);
    g_free(list);
}

void sd_bdii_freeServiceDetails(SDServiceDetails *details) {
    if (!details)
        return;

    g_free(details->name);
    g_free(details->type);
    g_free(details->endpoint);
    g_free(details->version);
    g_free(details->site);
    g_free(details->wsdl);
    g_free(details->administration);
    SD_freeVOList(details->vos);
    SD_freeServiceList(details->associatedServices);
    SD_freeServiceDataList(details->data);
    g_free(details);
}

void sd_bdii_freeServiceDetailsList(SDServiceDetailsList *servicedetailsList) {
    int i;
    
    if (!servicedetailsList)
        return;

    for (i = 0; i < servicedetailsList->numServiceDetails; i++)
        g_free(servicedetailsList->servicedetails[i]);
    g_free(servicedetailsList->servicedetails);
    g_free(servicedetailsList);
}


/* Convert SDServiceDetails to SDService */
SDService *details_to_service(const SDServiceDetails *details) {

    SDService *service = NULL;

    service = g_new0(SDService, 1);
    service->name = g_strdup(details->name);
    service->type = g_strdup(details->type);
    service->endpoint = g_strdup(details->endpoint);
    service->version = g_strdup(details->version);
    return service;
}


void add_sa_root(LDAP *ld, SDServiceDataList *list, const char *endpoint,
        SDException *exception) {

    LDAPMessage *reply = NULL;
    LDAPMessage *msg = NULL;
    GString *query = NULL;
    int ret = 0;
    g_setenv("G_SLICE","always-malloc",1);
    char *endpt = NULL; 

    if (!exception)
        exception = g_newa(SDException, 1);

    endpt = strdup(endpoint);
    char *host = get_host( endpt);

    query = g_string_new("(&" QUERY_OC_SA);
    g_string_append_printf(query, QUERY_SE, host);
    g_string_append(query, ")");


    /* returns an LDAP reply */
    ret = connect_search_ldap(default_search_base, query->str, sa_attrs, ld, exception, &reply, NULL, 0);
    if (ret != LDAP_SUCCESS) {
        SD_setException(exception, SDStatus_FAILURE,
                "add_sa_root: connect_search_ldap failed", NULL, 0);
        goto out;
    }

    if (ldap_count_entries(ld, reply) == 0) {		 
        goto out;
    }


    /* iterate the entries */
    for (msg = ldap_first_entry(ld, reply); msg; msg = ldap_next_entry(ld, msg)) {

        char *vo = get_single_value(ld, msg, ATTR_SA_ACBR);
        char *root = get_single_value(ld, msg, ATTR_SA_ROOT);

        if (!vo || !root){
	   if(vo)
	    g_free(vo);
	   if(root)
	    g_free(root);
	     
           continue;
	}

        char *value = strchr(root, ':');
        char *key = g_strdup_printf("%s:SEMountPoint", vo);
        g_free(vo);

        list->items = g_renew(SDServiceData, list->items, list->numItems + 1);
        list->items[list->numItems].key = key;

        if (NULL != value) {
            list->items[list->numItems].value = strdup(++value);
        } else {
            list->items[list->numItems].value = strdup(root);
        }
        ++list->numItems;

        g_free(root);
    }

out:
    if (reply)
        ldap_msgfree(reply);    
    if(query)
    	g_string_free(query, TRUE);
    if(endpt)
     free(endpt);

    return;

}



SDServiceDataList *query_service_data(LDAP *ld, LDAPMessage *imsg,
        const char *type, const char *endpoint, SDException *exception) {
    char *dn = NULL;
    SDServiceDataList *list = NULL;
    LDAPMessage *reply = NULL;
    LDAPMessage *msg = NULL;
    struct timeval timeout;
    int ret = 0;

    if (!exception)
        exception = g_newa(SDException, 1);

    dn = ldap_get_dn(ld, imsg);
    if (!dn) {
        SD_setException(exception, SDStatus_FAILURE,
                "cannot get DN", NULL, 0);
        return NULL;
    }


    timeout.tv_sec = get_timeout();
    timeout.tv_usec = 0;


    ret = ldap_search_ext_s(ld, dn, LDAP_SCOPE_SUBTREE, QUERY_DATA, data_attrs, 0, NULL, NULL,
            &timeout, 0, &reply);
    ldap_memfree(dn);

    if (ret != LDAP_SUCCESS) {
        report_query_error("LDAP search failed:  %s", ld, reply, ret, exception);
        goto error;
    }

    list = g_new0(SDServiceDataList, 1);

    for (msg = ldap_first_entry(ld, reply); msg; msg = ldap_next_entry(ld, msg)) {

        char *key, *value;

        key = get_single_value(ld, msg, ATTR_DATAKEY);
        value = get_single_value(ld, msg, ATTR_DATAVALUE);


        if (!key || !value) {

            g_free(key);
            g_free(value);
            continue;
        }

        list->items = g_renew(SDServiceData, list->items, list->numItems + 1);
        list->items[list->numItems].key = key;
        list->items[list->numItems].value = value;
        ++list->numItems;
    }


    if (reply)
        ldap_msgfree(reply);

    /* SPECIAL case for the SE/SRM case */
    if (strncasecmp(type, "srm", 3) == 0) {

        add_sa_root(ld, list, endpoint, exception);
    }

    return list;

error:
    sd_bdii_freeServiceDataList(list);
    if (reply)
        ldap_msgfree(reply);
    return NULL;
}


SDServiceList *do_list_query(const SDServiceDataList *data,
        const SDVOList *vos, LDAP *ld, GString *query, SDException *exception) {
    LDAPMessage *reply = NULL;
    LDAPMessage *msg = NULL;
    SDServiceList *list = NULL;
    int ret = 0;

    if (!exception)
        exception = g_newa(SDException, 1);

    struct timeval timeout;
    timeout.tv_sec = get_timeout();
    timeout.tv_usec = 0;

      ret = ldap_search_ext_s(ld, default_search_base, LDAP_SCOPE_SUBTREE, query->str, all_attrs, 0, NULL, NULL, &timeout, 0, &reply);
      if (query->str!= NULL){
        g_string_free(query, TRUE);
	query=NULL;	
	}
    if (ret != LDAP_SUCCESS) {
        report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
	goto out;
    }


    if (ldap_count_entries(ld, reply) == 0) {
        SD_setException(exception, SDStatus_FAILURE, "No entries found", NULL, 0);
        list = NULL;
        goto out;
    }

   
    list = g_new0(SDServiceList, 1);
    for (msg = ldap_first_entry(ld, reply); msg; msg = ldap_next_entry(ld, msg)) {

        SDService *service = NULL;

        service = g_new0(SDService, 1);
        ret = parse_service_attrs(service, ld, msg, exception);
        if (ret) {

            sd_bdii_freeService(service);
            sd_bdii_freeServiceList(list);
            list = NULL;
            goto out;
        }

        if (data) {


            SDServiceDataList *datalist = NULL;

            SD_setException(exception, SDStatus_SUCCESS, "Found", NULL, 0);
            datalist = query_service_data(ld, msg, service->type,
                    service->endpoint, exception);
            if (exception->status != SDStatus_SUCCESS)
                continue;

            /* Filter out non-matching entries */
            if (!match_data(datalist, data))
                continue;
        }

        if (vos) {

            /* Filter out entries not matching the requested VOs */
            if (!match_vos(ld, msg, vos))
                continue;
        }

        ret = add_to_list_bdii(list, service, exception);
        if (ret) {
            sd_bdii_freeService(service);
            sd_bdii_freeServiceList(list);
            list = NULL;
            goto out;
        }

    }
  
	
out:
    if (reply) 
        ldap_msgfree(reply);

    return list;
}



int parse_service_attrs(SDService *service, LDAP *ld, LDAPMessage *msg,
        SDException *exception) {

    service->name = get_single_value(ld, msg, ATTR_NAME);
    service->type = get_single_value(ld, msg, ATTR_TYPE);
    service->endpoint = get_single_value(ld, msg, ATTR_ENDPOINT);
    service->version = get_single_value(ld, msg, ATTR_VERSION);


    if (NULL == service->version) {
        char *major = get_single_value(ld, msg, ATTR_VERSION_MAJOR);
        char *minor = get_single_value(ld, msg, ATTR_VERSION_MINOR);
        char *patch = get_single_value(ld, msg, ATTR_VERSION_PATCH);
        service->version = (char *) malloc(20 * sizeof (char));
        if (major && minor && patch) {

            sprintf(service->version, "%s.%s.%s", major, minor, patch);
            g_free(major);
            g_free(minor);
            g_free(patch);
        } else if (major) {

            sprintf(service->version, "%s", major);
            g_free(major);
        } else {
            sprintf(service->version, "N/A");
        }
    }

    if (NULL == service->endpoint) {
        service->endpoint = get_single_value(ld, msg, ATTR_URI);
    }

    if (NULL == service->name) {
        service->name = get_single_value(ld, msg, ATTR_NM);
        if (NULL == service->name) {
            service->name = get_single_value(ld, msg, ATTR_URI);
        }
    }

    if (!service->name || !service->type || !service->endpoint) {
        char msg[1000];
        sprintf(
                msg,
                "Incomplete service definition: ServiceName:%s ServiceType:%s ServiceEndpoint:%s ServiceVersion:%s",
                !service->name ? "N/A" : service->name, !service->type ? "N/A"
                : service->type, !service->endpoint ? "N/A"
                : service->endpoint, !service->version ? "N/A"
                : service->version);       
        return -1;
    }

    /* the SRM fix (you can start laughing now) */
    if (strcmp(service->type, "srm_v1") == 0) {
        strcpy(service->type, "SRM");
    }

    return 0;
}


int match_data(const SDServiceDataList *list,
        const SDServiceDataList *requirements) {

    int i, j, gotit;

    /* Check that the service has all the specified key/value
     * pairs */
    for (i = 0; i < requirements->numItems; i++) {
        gotit = 0;

        for (j = 0; list && j < list->numItems; j++) {
            if (strcmp(requirements->items[i].key, list->items[j].key))
                continue;
            if (!strcmp(requirements->items[i].value, list->items[j].value)) {
                gotit = 1;
                break;
            }
        }

        if (!gotit)
            return FALSE;
    }

    return TRUE;
}

/* Returns TRUE if the service matches the given constraints, FALSE otherwise */
int match_service_bdii(SDServiceDetails *details, const char *type,
        const char *site, const SDVOList *vos) {

    int i, j, gotit;

    if (type && strcmp(details->type, type))
        return FALSE;
    if (site && details->site && strcmp(details->site, site))
        return FALSE;

    /* <volist> is not a required parameter. Treat an missing <volist>
     * as a "match all", an empty <volist> as a "match none" */
    if (vos && details->vos) {
        gotit = 0;

        /* A service that does not belong to any VOs matches the
         * empty list */
        if (!vos->numNames && !details->vos->numNames)
            gotit = 1;

        /* Check that at the intersection of vos and details->vos
         * is not empty */
        for (i = 0; i < vos->numNames && !gotit; i++) {
            for (j = 0; details->vos && j < details->vos->numNames; j++) {
                if (!strcmp(vos->names[i], details->vos->names[j])) {
                    gotit = 1;
                    break;
                }
            }
        }

        if (!gotit)
            return FALSE;
    }

    return TRUE;
}


int match_vos(LDAP *ld, LDAPMessage *msg, const SDVOList *vos) {

    struct berval **values = NULL;
    int i, j;
    char *VO = NULL;
    int isBase = 0;
    
    
      /* try the VO as it is sometimes used (accessrule) */
      values = ldap_get_values_len(ld, msg, ATTR_ACCESSRULE);
       if(!values)
       {
         values = ldap_get_values_len(ld, msg, ATTR_ACCESSRULEBASE);
         if(values)
		isBase = 1;
         else
               values = ldap_get_values_len(ld, msg, ATTR_VO);
       }
    

    /* if no VO is given at all, return TRUE, assume that the service is for all VOs */
    if (!values)
        return TRUE;

    for (i = 0; i < vos->numNames; i++) {
        for (j = 0; j < ldap_count_values_len(values); j++) {
            if (values && values[j] && values[j]->bv_val) {
               if(isBase == 0){
                if (0 == strcasecmp(vos->names[i], values[j]->bv_val)) {
                    ber_bvecfree(values);
                    return TRUE;
                }
               } else if(isBase == 1)
                {
                VO = get_vo(values[j]->bv_val);
                if (0 == strcasecmp(vos->names[i], VO)) {
                    ber_bvecfree(values);
                    return TRUE;
                }
         	}
            }
        }
    }

    if (values)
        ber_bvecfree(values);
    return FALSE;
}


/**
 * @description	fills a structure with attr as: name, type, endpoint, version */
int fill_service(LDAP *ld, const char *name, SDService *service, SDException *exception) {

    LDAPMessage *reply = NULL;
    char *query = NULL;   
    int ret = 0;

    if (!exception)
        exception = g_newa(SDException, 1);

    query = g_strdup_printf(QUERY_NAME, name, name);

    /* returns an LDAP reply */

    ret = connect_search_ldap(default_search_base, query, all_attrs, ld, exception, &reply, NULL, 0);
    if (ret != LDAP_SUCCESS) {
        SD_setException(exception, SDStatus_FAILURE, "fill_service query error", NULL, 0);
        /* clean up */
        if (reply)
            ldap_msgfree(reply);        
        if(query)
		g_free(query);
        return -1;
    }

    /* copy attr from reply to the SDService Structure */
    ret = parse_service_attrs(service, ld, reply, exception);

    /* clean up */
    if (reply)
        ldap_msgfree(reply);    
    if(query)
	g_free(query);

    return ret;
}


int add_associated_service(LDAP *ld, SDServiceDetails *details, const char *name,
        SDException *exception) {

    SDService *service=NULL, **tmp=NULL;
    int ret = 0;

    if (!exception)
        exception = g_newa(SDException, 1);

    service = g_new0(SDService, 1);
    ret = fill_service(ld, name, service, exception);
    if (ret) {
        /* We do not want to report an error if the pointed service does not exist */
        sd_bdii_freeService(service);
        return 0;
    }

    if (!details->associatedServices)
        details->associatedServices = g_new0(SDServiceList, 1);

    tmp = g_renew(SDService *, details->associatedServices->services,
            details->associatedServices->numServices + 1);
    tmp[details->associatedServices->numServices++] = service;
    details->associatedServices->services = tmp;
    return 0;
}


int add_vo(SDServiceDetails *details, const char *vo) {

    if (!details->vos)
        details->vos = g_new0(SDVOList, 1);

    details->vos->names
            = g_renew(char *, details->vos->names, details->vos->numNames + 1);
	    
    details->vos->names[details->vos->numNames++] = g_strdup(vo);
    return 0;
}


/* Add a service to a SDServiceList
 * add a structure like SDService to a list. */
int add_to_list_bdii(SDServiceList *list, SDService *service, SDException *ex) {

    SDService **tmp = NULL;

    if (!ex)
        ex = g_newa(SDException, 1);

    tmp = g_renew(SDService *, list->services, list->numServices + 1);
    if (!tmp) {
        SD_setException(ex, SDStatus_FAILURE, "Out of memory", NULL, 0);
        return -1;
    }

    tmp[list->numServices++] = service;
    list->services = tmp;
    return 0;
}


int parse_detail_attrs(SDServiceDetails *details, LDAP *ld, LDAPMessage *msg,
        SDException *exception) {

    struct berval **values = NULL;
    int i, ret;
    int isBase = 0;
    char *VO = NULL;

    SDService *service = NULL;
    service = g_new0(SDService, 1);

    if (!exception)
        exception = g_newa(SDException, 1);

    ret = parse_service_attrs(service, ld, msg, exception);
    if (ret) {
       if(service)
	  g_free(service);
        return ret;
    }

    details->name = service->name;
    details->type = service->type;
    details->endpoint = service->endpoint;
    details->version = service->version;
    details->wsdl = get_single_value(ld, msg, ATTR_WSDL);


    /* details->administration */
    /* ->site and ->associatedServices comes from ATTR_LINK */

    values = ldap_get_values_len(ld, msg, ATTR_LINK);
    ret = 0;
    for (i = 0; values && values[i] && values[i]->bv_val; i++) {

        char *name, *val;

        name = g_strdup(values[i]->bv_val);
        val = strchr(name, '=');
        if (!val) {

            g_free(name);
            continue;
        }

        *val++ = '\0';

        if (!strcasecmp(name, ATTR_SITE)) {

            g_free(details->site);
            details->site = g_strdup(val);
        } else if (!strcasecmp(name, ATTR_NAME)) {

            ret = add_associated_service(ld, details, val, exception);
            if (ret)
                break;
        }

        g_free(name);

    }


    ber_bvecfree(values);
    if (ret) {
        if(service)
	  g_free(service);    
        return ret;
    }

    if (!details->site) {
        /* it can be that the site is attached differently (old LCG model) */
        details->site = get_single_value(ld, msg, ATTR_HOSTINGORG);
    }

    values = ldap_get_values_len(ld, msg, ATTR_ACCESSRULE);
    if (!values) {
       values = ldap_get_values_len(ld, msg, ATTR_ACCESSRULEBASE);
      if(values){
        isBase = 1;
      }else{
        /* try the VO as it should be used */
        values = ldap_get_values_len(ld, msg, ATTR_VO);
      }
    }
    for (i = 0; values && values[i] && values[i]->bv_val; i++) {
        if(isBase == 0){
        ret = add_vo(details, values[i]->bv_val);
        if (ret) {
            break;
        }
    }
    else if (isBase == 1)
     {
        VO = get_vo(values[i]->bv_val);
        ret = add_vo(details, VO);
        if (ret) {
            break;
        }
     }
    }
    ber_bvecfree(values);

    if (ret) {
        if(service)
	  g_free(service);
        return ret;
    }
    SD_setException(exception, SDStatus_SUCCESS, "Success", NULL, 0);

   if(service)
       g_free(service);
    return 0;
}




int get_details(const char *q, const char *name, LDAP *ld, LDAPMessage **reply,
        SDException *exception) {
    char *query = NULL;
    struct timeval timeout;
    int ret;

    query = g_strdup_printf(q, name, name);

    timeout.tv_sec = get_timeout();
    timeout.tv_usec = 0;


    ret = ldap_search_ext_s(ld, default_search_base, LDAP_SCOPE_SUBTREE, query, all_attrs, 0, NULL, NULL,
            &timeout, 0, reply);
    g_free(query);
    if (ret != LDAP_SUCCESS) {
        report_query_error("LDAP search failed: %s", ld, *reply, ret, exception);
        return -1;
    }
    return 0;
}


SDServiceDetails *sd_bdii_getServiceDetails(const char *name,
        SDException *exception) {
    SDServiceDetails *details = NULL;
    LDAPMessage *reply = NULL; 
    int ret = 0;
    LDAP *ld = NULL;
    
    if ((ld = get_connection(exception, NULL, 0)) == NULL) {
        SD_setException(exception, SDStatus_FAILURE, "Cannot connect to BDII", NULL, 0);
        return NULL;
    }
     

    ret = get_details(QUERY_NAME, name, ld, &reply, exception);
    if (ret < 0) {
        goto error;
    }


    if (ldap_count_entries(ld, reply) == 0) {       
        goto error;
    }


    details = g_new0(SDServiceDetails, 1);
    ret = parse_detail_attrs(details, ld, reply, exception);
    if (ret) {
        goto error;
    }   

    details->data = query_service_data(ld, reply, details->type, details->endpoint, exception);
    if (exception->status != SDStatus_SUCCESS) {
        goto error;
    }


out:
    if (reply) 
        ldap_msgfree(reply);
    if(ld)
    	close_connection(ld);	
    return details;

error:
    if (details)
        sd_bdii_freeServiceDetails(details);
    details = NULL;
    goto out;


}

SDServiceList *sd_bdii_listAssociatedServices(const char *name,
        const char *type, const char *site, const SDVOList *vos,
        SDException *exception) {
    char *query = NULL;
    struct berval **values = NULL;
    SDServiceList *list = NULL;
    SDServiceDetailsList *detailslist = NULL;
    LDAPMessage *reply = NULL;
    int ret = 0, i = 0;
    LDAP *ld = NULL;

    if ((ld = get_connection(exception, NULL, 0)) == NULL) {
        SD_setException(exception, SDStatus_FAILURE, "Cannot connect to BDII", NULL, 0);
        return NULL;
    }
 

    if (!exception)
        exception = g_newa(SDException, 1);

    query = g_strdup_printf(QUERY_NAME, name, name);

    /* returns an LDAP reply */
    ret = connect_search_ldap(default_search_base, query, link_attrs, ld, exception, &reply, NULL, 0);
    if (ret != LDAP_SUCCESS) {
        report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
        goto out;
    }


    list = g_new0(SDServiceList, 1);
    detailslist = g_new0(SDServiceDetailsList, 1);
    if (!list || !detailslist) {
        SD_setException(exception, SDStatus_FAILURE, "Out of memory", NULL, 0);
        goto out;
    }

    values = ldap_get_values_len(ld, reply, ATTR_LINK);
    for (i = 0; values && values[i] && values[i]->bv_val; i++) {
        SDServiceDetails *details;
        SDService *service;

        /* We are only interested in service links */
        if (strncasecmp(values[i]->bv_val, ATTR_NAME, strlen(ATTR_NAME))
                || values[i]->bv_val[strlen(ATTR_NAME)] != '=')
            continue;

        /* A non-existant associated service is not an error */
        details = SD_getServiceDetails(values[i]->bv_val + strlen(
                ATTR_NAME) + 1, exception);
        if (!details)
            continue;

        /* Check if the service matches the requirements */
        if (!match_service_bdii(details, type, site, vos)) {
            sd_bdii_freeServiceDetails(details);
            continue;
        }

        service = details_to_service(details);
        ret = add_to_list_bdii(list, service, exception);
        if (ret) {
            sd_bdii_freeService(service);
            sd_bdii_freeServiceList(list);
            sd_bdii_freeServiceDetails(details);
            list = NULL;
            goto out;
        }
    }

    if (!list->numServices) {
        char buf[256];
        sd_bdii_freeServiceList(list);
        snprintf(buf, 255, "No services of type %s were found", type);
        SD_setException(exception, SDStatus_FAILURE, buf, NULL, 0);
        list = NULL;
        goto out;
    }
   
out:
    ber_bvecfree(values);
    if (reply) 
        ldap_msgfree(reply);
    if(query)
    	g_free(query);
    if(ld)
    	close_connection(ld);	

    return list;


}

SDServiceList *sd_bdii_listServicesByHost(const char *type, const char *host,
        const SDVOList *vos, SDException *exception) {

    g_setenv("G_SLICE","always-malloc",1);	
    GString *query = NULL;
    LDAP *ld = NULL;

    if (host == NULL) {
        return SD_listServicesByData(NULL, type, NULL, vos, exception);
    }

    if ((ld = get_connection(exception, NULL, 0)) == NULL) {
        SD_setException(exception, SDStatus_FAILURE, "Cannot connect to BDII", NULL, 0);
        return NULL;
    }
 

    /* GString is _very_ handy here :-) */
    query = g_string_new("(&" QUERY_OC_SERVICE);

    if (type) {
        /* this is the SRM hack so that both old and new types are looked for */
        if (strncasecmp(type, "srm", 3) == 0) {

            g_string_append(query, QUERY_SRM_TYPE);
        } else if (strncasecmp(type, "GSIFTP", 6) == 0) {

            g_string_append(query, QUERY_GSIFTP_TYPE);
        } else {

            g_string_append_printf(query, QUERY_TYPE, type);
        }
    }

    g_string_append_printf(query, QUERY_HOST, host, host);

    /* closing the and-ed subqueries */
    g_string_append(query, ")");

    SDServiceList *list = do_list_query(NULL, vos, ld, query, exception);
    if (list == NULL) {
        goto out;
    }

out:
    if(ld)
	close_connection(ld);	
    return list;    

}

SDServiceList *sd_bdii_listServicesByData(const SDServiceDataList *data,
        const char *type, const char *site, const SDVOList *vos,
        SDException *exception) {
	

    /* type, site and vos can be expressed as an LDAP filter. Data
     * key-value pairs however appear as child nodes in BDII so we have
     * to check them manually. */
    GString *query = NULL;
    LDAP *ld = NULL;
    g_setenv("G_SLICE","always-malloc",1); 
    
    if ((ld = get_connection(exception, NULL, 0)) == NULL) {
        SD_setException(exception, SDStatus_FAILURE, "Cannot connect to BDII", NULL, 0);
        return NULL;
    }
     

    if (!exception)
        exception = g_newa(SDException, 1);


    /* GString is _very_ handy here :-) */
    if (type || site)
        query = g_string_new("(&" QUERY_OC_SERVICE);
    else
        query = g_string_new(QUERY_OC_SERVICE);
	

    if (type) {
        if (strncasecmp(type, "srm", 3) == 0) {
            g_string_append(query, QUERY_SRM_TYPE);
        } else if (strncasecmp(type, "GSIFTP", 6) == 0) {
            g_string_append(query, QUERY_GSIFTP_TYPE);
        } else {
            g_string_append_printf(query, QUERY_TYPE, type);
        }
    }
    if (site)
        g_string_append_printf(query, QUERY_SITE, site);

    /* closing the and-ed subqueries */
    if (type || site)
        g_string_append(query, ")");


    SDServiceList *list = do_list_query(data, vos, ld, query, exception);
    if (list == NULL) {
        goto out;
    }

out:
    if(ld)
	close_connection(ld);	
    return list;
}



SDServiceList *sd_bdii_listServices(const char *type, const char *site, const SDVOList *vos, SDException *exception) {
    SDServiceList *result = NULL;
    SDVOList *tmpvos = NULL;

    if (!exception)
        exception = g_newa(SDException, 1);

    if (!site) {
        site = getenv(SDEnv_SITE);
        /* treat empty string as if it does not exist */
        if (site && (strlen(site) == 0)) site = NULL;
    }
    tmpvos = NULL;
    /* Let the user override the VO if the caller did not specify
     * one explicitely */
    if (!vos || !vos->numNames)
        tmpvos = checkVOEnv();


    result = SD_listServicesByData(NULL, type, site, tmpvos ? tmpvos : vos, exception);
 
    if (tmpvos)
        SD_freeVOList(tmpvos);
  
    return result;
}

SDService *sd_bdii_getService(const char *name, SDException *exception) {

    SDService *service = NULL;
    GString *query = NULL;
    
    LDAP *ld = NULL;
 
    /* GString is _very_ handy here :-) */
    if (!name) {
        SD_setException(exception, SDStatus_FAILURE,
                "Name not given in Service search", NULL, 0);
        return NULL;
    }

 
    if ((ld = get_connection(exception, NULL, 0)) == NULL) {
        SD_setException(exception, SDStatus_FAILURE, "Cannot connect to BDII", NULL, 0);
        return NULL;
    }
 
    query = g_string_sized_new(200);
    g_string_printf(query, QUERY_NAME, name, name);

    SDServiceList *list = do_list_query(NULL, NULL, ld, query, exception);
    if (list == NULL) {
        goto out;
    }
    if (list->numServices < 1) {
        goto out;
    }

    service = list->services[0];
    g_free(list->services);
    g_free(list);
   
out:
    if(ld)
	close_connection(ld);	
    return service;
}

SDServiceDataList *sd_bdii_getServiceData(const char *name,
        SDException *exception) {
    SDServiceDataList *list = NULL;

    LDAPMessage *reply = NULL;
    char *query = NULL;
    int ret = 0;

    LDAP *ld = NULL;
    if ((ld = get_connection(exception, NULL, 0)) == NULL) {
        SD_setException(exception, SDStatus_FAILURE, "Cannot connect to BDII", NULL, 0);
        return NULL;
    }


    /* query_service_data() returns the error code in the exception */
    if (!exception)
        exception = g_newa(SDException, 1);

    query = g_strdup_printf(QUERY_NAME, name, name);

    ret = connect_search_ldap(default_search_base, query, service_attrs, ld, exception, &reply, NULL, 0);
    if (ret != LDAP_SUCCESS) {
        report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
        list = NULL;
        goto out;
    }

    SDService *service;

    service = g_new0(SDService, 1);
    ret = parse_service_attrs(service, ld, reply, exception);
    if (ret) {
        sd_bdii_freeService(service);
        list = NULL;
        goto out;
    }

    SD_setException(exception, SDStatus_SUCCESS, "Success", NULL, 0);
    list = query_service_data(ld, reply, service->type, service->endpoint,
            exception);

out:
    if (reply) 
        ldap_msgfree(reply);
    if(ld)
	close_connection(ld);
    if(query)
	g_free(query);


    return list;
}

char *sd_bdii_getServiceSite(const char *name, SDException *exception) {

    char *query=NULL, *res = NULL;
    struct berval **values = NULL;
    LDAPMessage *reply = NULL;
    int ret=0, i=0;

    LDAP *ld = NULL;
    if ((ld = get_connection(exception, NULL, 0)) == NULL) {
        SD_setException(exception, SDStatus_FAILURE, "Cannot connect to BDII", NULL, 0);
        return NULL;
    }


    /* query_service_data() returns the error code in the exception */
    if (!exception)
        exception = g_newa(SDException, 1);

    query = g_strdup_printf(QUERY_NAME, name, name);

    ret = connect_search_ldap(default_search_base, query, site_attrs, ld, exception, &reply, NULL, 0);
    if (ret != LDAP_SUCCESS) {
        report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
        goto out;
    }
    
    values = ldap_get_values_len(ld, reply, ATTR_LINK);
    for (i = 0; values && values[i] && values[i]->bv_val; i++) {
        if (strncasecmp(values[i]->bv_val, ATTR_SITE, strlen(ATTR_SITE))
                || values[i]->bv_val[strlen(ATTR_SITE)] != '=')
            continue;
        res = strdup(values[i]->bv_val + strlen(ATTR_SITE) + 1);
    }
    ber_bvecfree(values);

out:
    if (reply) 
        ldap_msgfree(reply);
    if(query)
    	g_free(query);
    if(ld)
	close_connection(ld);
    return res;

}

char *sd_bdii_getServiceWSDL(const char *name, SDException *exception) {

    char *query=NULL, *res = NULL;
    LDAPMessage *reply = NULL;
    int ret = 0;

    LDAP *ld = NULL;
    if ((ld = get_connection(exception, NULL, 0)) == NULL) {
        SD_setException(exception, SDStatus_FAILURE, "Cannot connect to BDII", NULL, 0);
        return NULL;
    }


    if (!exception)
        exception = g_newa(SDException, 1);

    query = g_strdup_printf(QUERY_NAME, name, name);

    /* returns an LDAP reply */

    ret = connect_search_ldap(default_search_base, query, wsdl_attrs, ld, exception, &reply, NULL, 0);
    if (ret != LDAP_SUCCESS) {
        report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
        goto out;
    }

    res = get_single_value(ld, reply, ATTR_WSDL);
    if (!g_mem_is_system_malloc() && res) {
        char *tmp = strdup(res);
        g_free(res);
        res = tmp;
    }


out:
    if (reply) 
        ldap_msgfree(reply);
    if (query)
    	g_free(query);
    if(ld)
	close_connection(ld);
    return res;
}

char *sd_bdii_getServiceDataItem(const char *name, const char *key,
        SDException *exception) {

    char *query=NULL, *dn=NULL, *res = NULL;
    struct timeval timeout;
    LDAPMessage *reply = NULL;
    int ret = 0;

    LDAP *ld = NULL;
    if ((ld = get_connection(exception, NULL, 0)) == NULL) {
        SD_setException(exception, SDStatus_FAILURE, "Cannot connect to BDII", NULL, 0);
        return NULL;
    }


    if (!exception)
        exception = g_newa(SDException, 1);

    query = g_strdup_printf(QUERY_NAME, name, name);

    /* returns an LDAP reply */
    ret = connect_search_ldap(default_search_base, query, no_attrs, ld, exception, &reply, NULL, 0);
    if (ret != LDAP_SUCCESS) {       
        report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
	if(query)
		g_free(query);
        if (reply) 
                ldap_msgfree(reply);
        goto out;
    }
    else
    {
	if(query)
		g_free(query);    
    }
    
    dn = ldap_get_dn(ld, reply);
    if (!dn) {
        SD_setException(exception, SDStatus_FAILURE,
                "Failed to determine the DN", NULL, 0);
        if (reply) 
                ldap_msgfree(reply);		
        goto out;
    }

 
    query = g_strdup_printf(QUERY_DATAKEY, key);
    timeout.tv_sec = get_timeout();
    timeout.tv_usec = 0;

        if (reply) {
                ldap_msgfree(reply);
		reply  = NULL;
		}

    ret = ldap_search_ext_s(ld, dn, LDAP_SCOPE_SUBTREE, query, data_attrs, 0, NULL, NULL,
            &timeout, 0, &reply);
    g_free(query);
    ldap_memfree(dn);
    if (ret != LDAP_SUCCESS) {   
            if (reply) {
                ldap_msgfree(reply);
		reply  = NULL;
		}
        goto out;
    }

    res = get_single_value(ld, reply, ATTR_DATAVALUE);
    if (!g_mem_is_system_malloc() && res) {
        char *tmp = strdup(res);
        g_free(res);
        res = tmp;
    }

            if (reply) {
                ldap_msgfree(reply);
		reply  = NULL;
		}
 
out:
   if(ld)
	close_connection(ld);
    return res;
}


