/*                                                                -*- mode: c; c-basic-offset: 8;  -*-
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 ============================================================================
 Internal methods which are used for retrieving information from the cache file (services.xml)
 ============================================================================
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "ServiceDiscoveryIfce.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <pwd.h>

#include <glib.h>

/**********************************************************************
 * Constants
 */

/* Codes for the XML tags we recognize */
typedef enum {
    TAG_BEGIN,
    TAG_SERVICES,
    TAG_SERVICE,
    TAG_PARAMETERS,
    TAG_ENDPOINT,
    TAG_TYPE,
    TAG_VERSION,
    TAG_SITE,
    TAG_ADMIN,
    TAG_WSDL,
    TAG_VOLIST,
    TAG_VO,
    TAG_PARAM,
    TAG_ASSOCIATED,
    TAG_NAME
} tag_code;


/* Buffer size used when parsing the service file */
#define BUFLEN				2048

/*use hardcoced values for the possible path (gLite or EMI)*/
static char* filePaths[] = {"/opt/glite/etc/services.xml", "/usr/etc/services.xml"};

/**********************************************************************
 * Data type definitions
 */

typedef struct _parser_ctx parser_ctx;
typedef struct _tag_desc tag_desc;

struct _parser_ctx {
    /* The input file descriptor */
    int fd;
    /* The input buffer */
    char *buf;

    /* The XML parser */
    GMarkupParseContext *parser;
    /* Current XML tag */
    tag_code tag;
    /* Name of the file being parsed */
    const char *filename;

    /* The service being parsed */
    SDServiceDetails *service;
    /* Name of the current parameter */
    char *param_name;
    /* Last character data */
    char *char_data;
};

struct _tag_desc {
    /* Name of the tag */
    const char *name;
    /* Code of the tag */
    tag_code code;
    /* Code of the parent tag in which this tag is valid */
    tag_code parent;
    /* True if text data is allowed inside the tag */
    int text_valid;
    /* True if the tag may have attributes */
    int has_attrs;
};


/**********************************************************************
 * Prototypes
 */

static void parse_start_tag(GMarkupParseContext *parser, const char *name,
        const char **attr_names, const char **attr_values, void *ptr,
        GError **error);
static void parse_end_tag(GMarkupParseContext *parser, const char *name,
        void *ptr, GError **error);
static void parse_char_data(GMarkupParseContext *parser, const char *data,
        size_t len, void *ptr, GError **error);

/*internal ifce*/
SDService *sd_file_getService(const char *name, SDException *exception);
SDServiceDetails *sd_file_getServiceDetails(const char *name,
        SDException *exception);
SDServiceDataList *sd_file_getServiceData(const char *name,
        SDException *exception);
SDServiceList *sd_file_listAssociatedServices(const char *name,
        const char *type, const char *site, const SDVOList *vos,
        SDException *exception);
SDServiceList *sd_file_listServicesByData(const SDServiceDataList *data,
        const char *type, const char *site, const SDVOList *vos,
        SDException *exception);


void sd_file_freeServiceDataList(SDServiceDataList *serviceDataList);
void sd_file_freeService(SDService *service);
void sd_file_freeServiceList(SDServiceList *serviceList);
void sd_file_freeServiceDetailsList(SDServiceDetailsList *servicedetailsList);
void sd_file_freeVOList(SDVOList *vos);
void sd_file_freeServiceDetails(SDServiceDetails *details);
void sd_file_sortbyVO(SDServiceDetailsList *servicedetailsList, const SDVOList *vos);

#ifndef __GNUC__
#define __attribute__(x)
#endif


/**********************************************************************
 * Global variables
 */

/* Description of the XML tags we recognize */
static const tag_desc tag_table[] ={
    { "services", TAG_SERVICES, TAG_BEGIN, 0, 0},
    { "service", TAG_SERVICE, TAG_SERVICES, 0, 1},
    { "parameters", TAG_PARAMETERS, TAG_SERVICE, 0, 0},
    { "endpoint", TAG_ENDPOINT, TAG_PARAMETERS, 1, 0},
    { "type", TAG_TYPE, TAG_PARAMETERS, 1, 0},
    { "version", TAG_VERSION, TAG_PARAMETERS, 1, 0},
    { "site", TAG_SITE, TAG_PARAMETERS, 1, 0},
    { "wsdl", TAG_WSDL, TAG_PARAMETERS, 1, 0},
    { "administration", TAG_ADMIN, TAG_PARAMETERS, 1, 0},
    { "volist", TAG_VOLIST, TAG_PARAMETERS, 0, 0},
    { "vo", TAG_VO, TAG_VOLIST, 1, 0},
    { "param", TAG_PARAM, TAG_PARAMETERS, 1, 1},
    { "associatedservices", TAG_ASSOCIATED, TAG_SERVICE, 0, 0},
    { "name", TAG_NAME, TAG_ASSOCIATED, 1, 0}
};

/* GMarkup parser description */
static const GMarkupParser services_parser ={
    .start_element = parse_start_tag,
    .end_element = parse_end_tag,
    .text = parse_char_data
};

/* List of loaded services */
static GSList *services;

/* Name of the service file */
static char *service_file;
/* Last state of the service file */
static struct stat service_stat;



/**********************************************************************
 * Misc. utilities
 */

/* Set the error string in the exception */
void set_error(SDException *ex, const char *fmt, ...) {
    char buf[1024];
    va_list ap;

    if (!ex)
        return;

    va_start(ap, fmt);
    vsnprintf(buf, sizeof (buf), fmt, ap);
    va_end(ap);

    ex->reason = g_strdup(buf);
    ex->status = SDStatus_FAILURE;
}

/* Find a service by name */
SDServiceDetails *find_service(const char *name) {
    if (!name)
        return NULL;

    SDServiceDetails *details;
    GSList *l;

    for (l = services; l; l = l->next) {
        details = l->data;
        if (!strcmp(details->name, name))
            return details;
    }
    return NULL;
}

/* Add a service to a SDServiceList */
int add_to_list(SDServiceList *list, SDService *service, SDException *ex) {
    SDService **tmp;

    tmp = g_renew(SDService *, list->services, list->numServices + 1);
    if (!tmp) {
        set_error(ex, "Out of memory");
        return -1;
    }

    tmp[list->numServices++] = service;
    list->services = tmp;
    return 0;
}

/* Add a service to a SDServiceDetailsList */
int add_to_details_list(SDServiceDetailsList *list, SDServiceDetails *servicedetails, SDException *ex) {
    SDServiceDetails **tmp;

    tmp = g_renew(SDServiceDetails *, list->servicedetails, list->numServiceDetails + 1);
    if (!tmp) {
        set_error(ex, "Out of memory");
        return -1;
    }

    tmp[list->numServiceDetails++] = servicedetails;
    list->servicedetails = tmp;
    return 0;
}

/* Returns TRUE if the service matches the given constraints, FALSE otherwise */
int match_service(SDServiceDetails *details, const char *type,
        const char *site, const SDVOList *vos, const SDServiceDataList *data) {
    int i, j, gotit;

    if (type && strcmp(details->type, type))
        return 0;
    if (site && details->site && strcmp(details->site, site))
        return 0;

    /* <volist> is not a required parameter. Treat an missing <volist>
     * as a "match all", an empty <volist> as a "match none" */
    if (vos && details->vos) {
        gotit = 0;

        /* A service that does not belong to any VOs matches the
         * empty list */
        if (!vos->numNames && !details->vos->numNames)
            gotit = 1;

        /* Check that at the intersection of vos and details->vos
         * is not empty */
        for (i = 0; i < vos->numNames && !gotit; i++) {
            for (j = 0; details->vos && j < details->vos->numNames;
                    j++) {
                if (!strcmp(vos->names[i],
                        details->vos->names[j])) {
                    gotit = 1;
                    break;
                }
            }
        }

        if (!gotit)
            return 0;
    }

    if (data) {
        /* Check that the service has all the specified key/value
         * pairs */
        for (i = 0; i < data->numItems; i++) {
            gotit = 0;

            for (j = 0; details->data && j < details->data->numItems;
                    j++) {
                if (strcmp(data->items[i].key,
                        details->data->items[j].key))
                    continue;
                if (!strcmp(data->items[i].value,
                        details->data->items[j].value)) {
                    gotit = 1;
                    break;
                }
            }

            if (!gotit)
                return 0;
        }
    }

    return 1;
}

/* Make a deep copy of a SDServiceDataList */
SDServiceDataList *clone_datalist(SDServiceDataList *orig) {
    SDServiceDataList *clone;
    int i;

    if (!orig || !orig->numItems)
        return NULL;

    clone = g_new0(SDServiceDataList, 1);
    if (!clone) {

        sd_file_freeServiceDataList(clone);
        return NULL;
    }

    clone->items = g_new0(SDServiceData, orig->numItems);
    if (orig->numItems && !clone->items) {

        sd_file_freeServiceDataList(clone);
        return NULL;
    }

    clone->numItems = orig->numItems;
    for (i = 0; i < clone->numItems; i++) {
        clone->items[i].key = g_strdup(orig->items[i].key);
        clone->items[i].value = g_strdup(orig->items[i].value);
        if (!clone->items[i].key || !clone->items[i].value) {

            sd_file_freeServiceDataList(clone);
            return NULL;
        }
    }

    return clone;
}

/* Make a deep copy of a SDServiceList */
SDServiceList *clone_servicelist(SDServiceList *orig) {
    SDServiceList *clone;
    int i;

    if (!orig || !orig->numServices)
        return NULL;

    clone = g_new0(SDServiceList, 1);
    if (!clone)
        return NULL;

    clone->services = g_new0(SDService *, orig->numServices);
    if (orig->numServices && !clone->services) {

        sd_file_freeServiceList(clone);
        return NULL;
    }

    clone->numServices = orig->numServices;
    for (i = 0; i < clone->numServices; i++) {
        clone->services[i] = g_new0(SDService, 1);
        if (!clone->services[i]) {

            sd_file_freeServiceList(clone);
            return NULL;
        }
        clone->services[i]->name = g_strdup(orig->services[i]->name);
        clone->services[i]->type = g_strdup(orig->services[i]->type);
        clone->services[i]->endpoint =
                g_strdup(orig->services[i]->endpoint);
        clone->services[i]->version =
                g_strdup(orig->services[i]->version);
        if (!clone->services[i]->name || !clone->services[i]->type ||
                !clone->services[i]->endpoint ||
                !clone->services[i]->version) {

            sd_file_freeServiceList(clone);
            return NULL;
        }
    }

    return clone;
}

/* Make a deep copy of a SDServiceDetails */
SDServiceDetails *clone_details(SDServiceDetails *orig) {
    SDServiceDetails *clone;

    clone = g_new0(SDServiceDetails, 1);
    if (!clone)
        return NULL;

    clone->name = g_strdup(orig->name);
    clone->type = g_strdup(orig->type);
    clone->endpoint = g_strdup(orig->endpoint);
    clone->version = g_strdup(orig->version);
    if (orig->site)
        clone->site = g_strdup(orig->site);
    if (orig->administration)
        clone->administration = g_strdup(orig->administration);
    if (orig->wsdl)
        clone->wsdl = g_strdup(orig->wsdl);
    if (!clone->name || !clone->type || !clone->endpoint || !clone->version ||
            (orig->site && !clone->site) ||
            (orig->wsdl && !clone->wsdl) ||
            (orig->administration && !clone->administration)) {

        sd_file_freeServiceDetails(clone);
        return NULL;
    }

    if (orig->vos) {
        int i;

        clone->vos = g_new0(SDVOList, 1);
        if (!clone->vos) {

            sd_file_freeServiceDetails(clone);
            return NULL;
        }

        clone->vos->names = g_new(char *, orig->vos->numNames);
        if (orig->vos->numNames && !clone->vos->names) {

            sd_file_freeServiceDetails(clone);
            return NULL;
        }

        clone->vos->numNames = orig->vos->numNames;
        for (i = 0; i < clone->vos->numNames; i++) {
            clone->vos->names[i] = g_strdup(orig->vos->names[i]);
            if (!clone->vos->names[i]) {
                clone->vos->numNames = i;
                sd_file_freeServiceDetails(clone);
                return NULL;
            }
        }
    }

    clone->data = clone_datalist(orig->data);
    if (orig->data && orig->data->numItems && !clone->data) {

        sd_file_freeServiceDetails(clone);
        return NULL;
    }

    clone->associatedServices = clone_servicelist(orig->associatedServices);
    if (orig->associatedServices &&
            orig->associatedServices->numServices &&
            !clone->associatedServices) {

        sd_file_freeServiceDetails(clone);
        return NULL;
    }

    return clone;
}

/* Remove all cached data */
void kill_cache(void) {
    while (services) {
        GSList * s = services;

        sd_file_freeServiceDetails(s->data);
        s->data = NULL;
        services = g_slist_remove_link(services, s);
        g_slist_free_1(s);
    }
}

/* Convert SDServiceDetails to SDService */
static SDService *details_to_service(const SDServiceDetails *details,
        SDException *exception) {
    SDService *service;

    service = g_new0(SDService, 1);
    if (!service) {
        set_error(exception, "Out of memory");
        return NULL;
    }

    service->name = g_strdup(details->name);
    service->type = g_strdup(details->type);
    service->endpoint = g_strdup(details->endpoint);
    service->version = g_strdup(details->version);
    if (!service->name || !service->type || !service->endpoint ||
            !service->version) {
        set_error(exception, "Out of memory");
        sd_file_freeService(service);
        return NULL;
    }

    return service;
}

/**********************************************************************
 * XML parser routines
 */

/* Terminate the parser with an error message */
void parse_error(parser_ctx *ctx, GError **error, const char *fmt, ...) {
    int line, chpos;
    char buf[1024];
    va_list ap;

    g_markup_parse_context_get_position(ctx->parser, &line, &chpos);
    va_start(ap, fmt);
    snprintf(buf, sizeof (buf), "Parsing error in %s line %d at %d: ",
            ctx->filename, line, chpos);
    vsnprintf(buf + strlen(buf), sizeof (buf) - strlen(buf), fmt, ap);
    va_end(ap);

    g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
            "%s", buf);
}

/* Allocate a new parser context */
parser_ctx *parser_ctx_new(const char *filename, SDException *ex) {
    parser_ctx *ctx;

    ctx = g_new0(parser_ctx, 1);
    if (!ctx) {
        set_error(ex, "Out of memory");
        return NULL;
    }

    ctx->buf = g_malloc(BUFLEN);
    if (!ctx->buf) {
        set_error(ex, "Failed to allocate the read buffer");
        g_free(ctx);
        return NULL;
    }

    ctx->tag = TAG_BEGIN;
    ctx->filename = filename;
    ctx->fd = open(ctx->filename, O_RDONLY);
    if (ctx->fd == -1) {
        g_free(ctx->buf);
        g_free(ctx);
        return NULL;
    }

    return ctx;
}

/* Free resources allocated to the parsing context */
void parser_ctx_free(parser_ctx *ctx) {


    if (ctx->fd != -1)
        close(ctx->fd);
    g_free(ctx->buf);
    g_markup_parse_context_free(ctx->parser);

    sd_file_freeServiceDetails(ctx->service);
    g_free(ctx->param_name);
    g_free(ctx->char_data);
    g_free(ctx);


}

/* Look up a tag descriptor by tag name */
const tag_desc *find_tag(const char *name) {
    unsigned i;

    for (i = 0; i < sizeof (tag_table) / sizeof (tag_table[0]); i++)
        if (!strcmp(tag_table[i].name, name))
            return tag_table + i;
    return NULL;
}

/* Look up a tag descriptor by tag code */
const tag_desc *find_tag_by_code(tag_code code) {
    unsigned i;

    for (i = 0; i < sizeof (tag_table) / sizeof (tag_table[0]); i++)
        if (tag_table[i].code == code)
            return tag_table + i;
    return NULL;
}

/* Process an opening <service> tag */
void process_service(parser_ctx *ctx, const char **attr_names,
        const char **attr_values, GError **error) {
    int i;

    /* Check if we already have a service */
    if (ctx->service) {
        parse_error(ctx, error, "Service is already open?!?");
        return;
    }

    /* Find the name attribute, ignore everything else */
    for (i = 0; attr_names && attr_names[i]; i++)
        if (!strcmp(attr_names[i], "name"))
            break;

    if (!attr_names || !attr_names[i]) {
        parse_error(ctx, error, "Service name is missing");
        return;
    }

    ctx->service = find_service(attr_values[i]);
    if (ctx->service) {
        /* Avoid it being freed twice */
        ctx->service = NULL;
        parse_error(ctx, error, "Duplicate service name");
        return;
    }

    /* Allocate a new service detail structure that will be filled
     * in by the parser */
    ctx->service = g_new0(SDServiceDetails, 1);
    if (!ctx->service) {
        parse_error(ctx, error, "Out of memory");
        return;
    }

    ctx->service->name = g_strdup(attr_values[i]);
    if (!ctx->service->name) {
        parse_error(ctx, error, "Out of memory");
        g_free(ctx->service);
        ctx->service = NULL;
        return;
    }
}

/* Add the just processed service to our internal list */
void commit_service(parser_ctx *ctx, GError **error) {
    if (!ctx->service || !ctx->service->name || !ctx->service->type ||
            !ctx->service->endpoint || !ctx->service->version) {
        parse_error(ctx, error, "Incomplete service definition");
        return;
    }

    services = g_slist_append(services, ctx->service);
    ctx->service = NULL;
}

/* Process an opening <param> tag */
void process_param(parser_ctx *ctx, const char **attr_names,
        const char **attr_values, GError **error) {
    int i;

    /* Find the name attribute, ignore everything else */
    for (i = 0; attr_names && attr_names[i]; i++)
        if (!strcmp(attr_names[i], "name"))
            break;

    if (!attr_names || !attr_names[i]) {
        parse_error(ctx, error, "Parameter name is missing");
        return;
    }

    g_free(ctx->param_name);
    ctx->param_name = g_strdup(attr_values[i]);
    if (!ctx->param_name)
        parse_error(ctx, error, "Out of memory");
}

/* Add a parameter to a service */
void commit_param(parser_ctx *ctx, GError **error) {
    SDServiceData *list;

    if (!ctx->service->data) {
        ctx->service->data = g_new0(SDServiceDataList, 1);
        if (!ctx->service->data) {
            parse_error(ctx, error, "Out of memory");
            return;
        }
    }

    list = g_renew(SDServiceData, ctx->service->data->items,
            ctx->service->data->numItems + 1);
    if (!list) {
        parse_error(ctx, error, "Out of memory");
        return;
    }

    list[ctx->service->data->numItems].key = ctx->param_name;
    list[ctx->service->data->numItems].value = ctx->char_data;
    ctx->param_name = NULL;
    ctx->char_data = NULL;

    ctx->service->data->items = list;
    ctx->service->data->numItems++;

    /* API says value cannot be NULL */
    if (!list[ctx->service->data->numItems - 1].value)
        list[ctx->service->data->numItems - 1].value = g_strdup("");
}

/* Add a VO to a service */
void commit_vo(parser_ctx *ctx, GError **error) {
    char **tmp;

    if (!ctx->char_data) {
        parse_error(ctx, error, "VO name cannot be empty");
        return;
    }

    tmp = g_renew(char *, ctx->service->vos->names,
            ctx->service->vos->numNames + 1);
    if (!tmp) {
        parse_error(ctx, error, "Out of memory");
        return;
    }

    tmp[ctx->service->vos->numNames++] = ctx->char_data;
    ctx->char_data = NULL;
    ctx->service->vos->names = tmp;
}

/* Add an associated service */
void commit_name(parser_ctx *ctx, GError **error) {
    SDService **list, *service;

    if (!ctx->char_data) {
        parse_error(ctx, error, "The name of the associated service "
                "must not be empty");
        return;
    }

    if (!ctx->service->associatedServices) {
        ctx->service->associatedServices = g_new0(SDServiceList, 1);
        if (!ctx->service->associatedServices) {
            parse_error(ctx, error, "Out of memory");
            return;
        }
    }

    service = g_new0(SDService, 1);
    if (!service) {
        parse_error(ctx, error, "Out of memory");
        return;
    }

    /* We only fill in the name field, the rest will be filled in by
     * parse_file() after all the services have been read, so it is
     * not neccessary to order services according to the relations */
    service->name = ctx->char_data;
    ctx->char_data = NULL;

    list = g_renew(SDService *, ctx->service->associatedServices->services,
            ctx->service->associatedServices->numServices + 1);
    if (!list) {
        parse_error(ctx, error, "Out of memory");
        g_free(service->name);
        g_free(service);
        return;
    }
    list[ctx->service->associatedServices->numServices++] = service;
    ctx->service->associatedServices->services = list;
}

/* XML callback for start tags */
void parse_start_tag(GMarkupParseContext *parser G_GNUC_UNUSED,
        const char *name, const char **attr_names, const char **attr_values,
        void *ptr, GError **error) {
    parser_ctx *ctx = ptr;
    const tag_desc *tag;

    tag = find_tag(name);
    if (!tag) {
        parse_error(ctx, error, "Unknown tag %s", name);
        return;
    }

    /* Check that new tag is allowed as a child of the current tag */
    if (tag->parent != ctx->tag) {
        parse_error(ctx, error, "Tag %s is not valid here", name);
        return;
    }

    if (attr_names && attr_names[0] && !tag->has_attrs) {
        parse_error(ctx, error, "Tag %s cannot have attributes", name);
        return;
    }

    /* Process the individual tags. If the tag does not have attributes, we
     * do not have to do anything here. */
    switch (tag->code) {
        case TAG_BEGIN:
        case TAG_SERVICES:
        case TAG_PARAMETERS:
        case TAG_ENDPOINT:
        case TAG_TYPE:
        case TAG_VERSION:
        case TAG_SITE:
        case TAG_ADMIN:
        case TAG_WSDL:
        case TAG_VO:
        case TAG_ASSOCIATED:
        case TAG_NAME:
            break;
        case TAG_VOLIST:
            /* Allocate the VO list here so we can differentiate
             * between an empty and a missing list */
            ctx->service->vos = g_new0(SDVOList, 1);
            if (!ctx->service->vos) {
                parse_error(ctx, error, "Out of memory");
                return;
            }
            break;
        case TAG_SERVICE:
            process_service(ctx, attr_names, attr_values, error);
            break;
        case TAG_PARAM:
            process_param(ctx, attr_names, attr_values, error);
            break;
    }

    /* Invalidate any left-over character data */
    g_free(ctx->char_data);
    ctx->char_data = NULL;

    /* Set the current tag in the context */
    ctx->tag = tag->code;
}

/* Set the given variable to the character data inside the last tag */
void set_var(parser_ctx *ctx, char **var, int mayempty, GError **error) {
    if (!mayempty && !ctx->char_data) {
        const tag_desc *tag;

        tag = find_tag_by_code(ctx->tag);
        parse_error(ctx, error, "Tag <%s> cannot be empty", tag->name);
        return;
    }

    g_free(*var);
    *var = ctx->char_data;
    ctx->char_data = NULL;
}

/* XML callback for end tags */
void parse_end_tag(GMarkupParseContext *parser G_GNUC_UNUSED,
        const char *name, void *ptr, GError **error) {
    parser_ctx *ctx = ptr;
    const tag_desc *tag;

    tag = find_tag(name);
    if (!tag) {
        parse_error(ctx, error, "Unknown tag %s", name);
        return;
    }

    switch (tag->code) {
        case TAG_BEGIN:
        case TAG_SERVICES:
        case TAG_PARAMETERS:
        case TAG_VOLIST:
        case TAG_ASSOCIATED:
            break;
        case TAG_ENDPOINT:
            set_var(ctx, &ctx->service->endpoint, 0, error);
            break;
        case TAG_TYPE:
            set_var(ctx, &ctx->service->type, 0, error);
            break;
        case TAG_VERSION:
            set_var(ctx, &ctx->service->version, 0, error);
            break;
        case TAG_SITE:
            set_var(ctx, &ctx->service->site, 1, error);
            break;
        case TAG_WSDL:
            set_var(ctx, &ctx->service->wsdl, 1, error);
            break;
        case TAG_ADMIN:
            set_var(ctx, &ctx->service->administration, 1, error);
            break;
        case TAG_SERVICE:
            commit_service(ctx, error);
            break;
        case TAG_PARAM:
            commit_param(ctx, error);
            break;
        case TAG_VO:
            commit_vo(ctx, error);
            break;
        case TAG_NAME:
            commit_name(ctx, error);
            break;
    }

    /* Set the current tag to the parent's code */
    ctx->tag = tag->parent;
}

/* Callback for character data */
void parse_char_data(GMarkupParseContext *parser G_GNUC_UNUSED,
        const char *data, size_t len, void *ptr, GError **error) {
    parser_ctx *ctx = ptr;
    const tag_desc *tag;

    tag = find_tag_by_code(ctx->tag);
    if (!tag) {
        parse_error(ctx, error, "Illegal parser state");
        return;
    }

    ctx->char_data = g_realloc(ctx->char_data, len + 1);
    if (!ctx->char_data) {
        parse_error(ctx, error, "Out of memory");
        return;
    }

    /* Note: data is not 0-terminated! */
    memcpy(ctx->char_data, data, len);
    ctx->char_data[len] = '\0';

    if (!tag->text_valid) {
        /* Allow whitespace anywhere in the document */
        if (strspn(ctx->char_data, " \t\n\r") == strlen(ctx->char_data))
            return;
        parse_error(ctx, error, "Character data is not allowed here");
        return;
    }
}

/* Fill in missing information in the associatedServices list */
void complete_service(void *data, void *ptr) {
    SDServiceDetails *details = data;
    SDException *ex = ptr;
    SDServiceList *list;
    int j;

    if (ex->status != SDStatus_SUCCESS)
        return;

    list = details->associatedServices;
    for (j = 0; list && j < list->numServices; j++) {
        SDService *assoc = list->services[j];
        SDServiceDetails *service;

        service = find_service(assoc->name);
        if (!service) {
            set_error(ex, "Service %s refers to unknown "
                    "service %s", details->name,
                    assoc->name);
            return;
        }

        assoc->type = g_strdup(service->type);
        assoc->endpoint = g_strdup(service->endpoint);
        assoc->version = g_strdup(service->version);
        if (!assoc->type || !assoc->endpoint ||
                !assoc->version) {
            set_error(ex, "Out of memory");
            return;
        }
    }
}

int parse_file(const char *filename, SDException *ex) {
    parser_ctx *ctx;
    GError *error;
    int ret;

    /* Initialize the parser context */
    ctx = parser_ctx_new(filename, ex);
    if (!ctx)
        return ex->status == SDStatus_SUCCESS ? 0 : -1;

    ctx->parser = g_markup_parse_context_new(&services_parser, 0, ctx,
            NULL);
    if (!ctx->parser) {
        set_error(ex, "Failed to create XML parser");
        parser_ctx_free(ctx);
        return -1;
    }

    error = NULL;
    for (;;) {
        ret = read(ctx->fd, ctx->buf, BUFLEN);
        if (ret < 0) {
            set_error(ex, "Error reading %s: %s", ctx->filename,
                    strerror(errno));
            parser_ctx_free(ctx);
            return -1;
        }

        if (!g_markup_parse_context_parse(ctx->parser, ctx->buf, ret,
                &error)) {
            if (error)
                set_error(ex, "%s", error->message);
            else
                set_error(ex, "Unknown error parsing %s",
                    ctx->filename);
            g_error_free(error);
            parser_ctx_free(ctx);
            return -1;
        }

        if (!ret)
            break;
    }

    /* Check if the document was complete */
    if (!g_markup_parse_context_end_parse(ctx->parser, &error)) {
        if (error)
            set_error(ex, "%s", error->message);
        else
            set_error(ex, "Unknown error parsing %s",
                ctx->filename);
        g_error_free(error);
        parser_ctx_free(ctx);
        return -1;
    }

    parser_ctx_free(ctx);

    /* Now complete the service associations */
    g_slist_foreach(services, complete_service, ex);
    if (ex->status != SDStatus_SUCCESS)
        return -1;
    return 0;
}

int check_init_servicelist(SDException *exception) {
    int ret;

    if (exception)
        memset(exception, 0, sizeof (*exception));


    if (!service_file) {
        char *filename = NULL;
        char * pPath = getenv("GLITE_SD_SERVICES_XML");
        if (pPath != NULL) {
            if (access(pPath, R_OK) == 0) {
                filename = pPath;
            }
        }

        if (access(filePaths[0], R_OK) == 0) {
            filename = filePaths[0];
        } else if (access(filePaths[1], R_OK) == 0) {
            filename = filePaths[1];
        }
        /* Not having a service list is not an error */
        if (!filename)
            return 0;


        ret = stat(filename, &service_stat);
        if (ret) {
            return 0;
        }

        service_file = filename;
    } else {
        struct stat st;

        ret = stat(service_file, &st);

        /* If the service file is gone, keep the cached information */
        if (ret)
            return 0;

        /* If the file has not changed, do not parse it again */
        if (st.st_dev == service_stat.st_dev &&
                st.st_ino == service_stat.st_ino &&
                st.st_size == service_stat.st_size &&
                st.st_mtime == service_stat.st_mtime)
            return 0;

        /* The service file has changed, drop the cached data and
         * re-parse the file */
        kill_cache();
    }

    ret = parse_file(service_file, exception);
    if (0 == ret) {

        ret = stat(service_file, &service_stat);
    }
    return ret;
}

SDService *sd_file_getService(const char *name, SDException *exception) {
    SDServiceDetails *details;

    if (check_init_servicelist(exception))
        return NULL;

    details = find_service(name);
    if (!details) {
        set_error(exception, "Unknown service name");
        return NULL;
    }

    return details_to_service(details, exception);
}

SDServiceDetails *sd_file_getServiceDetails(const char *name,
        SDException *exception) {
    SDServiceDetails *orig, *clone;

    if (check_init_servicelist(exception))
        return NULL;

    orig = find_service(name);
    if (!orig) {
        set_error(exception, "Unknown service name");
        return NULL;
    }

    clone = clone_details(orig);
    if (!clone)
        set_error(exception, "Out of memory");

    return clone;
}

SDServiceDataList *sd_file_getServiceData(const char *name,
        SDException *exception) {
    SDServiceDetails *service;
    SDServiceDataList *data;

    if (check_init_servicelist(exception))
        return NULL;

    service = find_service(name);
    if (!service) {
        set_error(exception, "Unknown service name");
        return NULL;
    }

    data = clone_datalist(service->data);
    if (service->data && service->data->numItems && !data)
        set_error(exception, "Out of memory");
    return data;
}

SDServiceList *sd_file_listAssociatedServices(const char *name,
        const char *type, const char *site, const SDVOList *vos,
        SDException *exception) {
    SDServiceList *list, *origlist;
    SDServiceDetailsList *detailslist;
    SDServiceDetails *service;
    int i;

    if (check_init_servicelist(exception))
        return NULL;

    service = find_service(name);
    if (!service) {
        set_error(exception, "Unknown service name");
        return NULL;
    }

    list = g_new0(SDServiceList, 1);
    detailslist = g_new0(SDServiceDetailsList, 1);
    if (!detailslist || !list) {
        set_error(exception, "Out of memory");
        return NULL;
    }

    origlist = service->associatedServices;
    for (i = 0; origlist && i < origlist->numServices; i++) {
        SDServiceDetails *details;
        int ret;

        details = find_service(origlist->services[i]->name);
        if (!details)
            continue;
        if (!match_service(details, type, site, vos, NULL))
            continue;

        ret = add_to_details_list(detailslist, details, exception);
        if (ret) {
            sd_file_freeServiceDetailsList(detailslist);
            return NULL;
        }
    }

    /* Sort detailslist by VOMS */
    sd_file_sortbyVO(detailslist, vos);

    for (i = 0; i < detailslist->numServiceDetails; i++) {
        SDService *aserv;
        int ret;

        aserv = details_to_service(detailslist->servicedetails[i], exception);
        if (!aserv) {
            SD_freeException(exception);
            continue;
        }

        ret = add_to_list(list, aserv, exception);
        if (ret) {
            sd_file_freeService(aserv);
            sd_file_freeServiceList(list);
            sd_file_freeServiceDetailsList(detailslist);
            return NULL;
        }
    }

    sd_file_freeServiceDetailsList(detailslist);

    if (!list->numServices) {
        g_free(list);
        set_error(exception, "No services of type %s were found", type);
        return NULL;
    }

    return list;
}

SDServiceList *sd_file_listServicesByData(const SDServiceDataList *data,
        const char *type, const char *site, const SDVOList *vos,
        SDException *exception) {
    SDServiceList *list;
    SDServiceDetailsList *detailslist;
    GSList *l;
    int i;

    if (check_init_servicelist(exception))
        return NULL;

    list = g_new0(SDServiceList, 1);
    detailslist = g_new0(SDServiceDetailsList, 1);
    if (!list || !detailslist) {
        set_error(exception, "Out of memory");
        return NULL;
    }

    for (l = services; l; l = l->next) {
        SDServiceDetails *details = l->data;
        int ret;

        if (!match_service(details, type, site, vos, data))
            continue;

        ret = add_to_details_list(detailslist, details, exception);
        if (ret) {
            sd_file_freeServiceDetailsList(detailslist);
            return NULL;
        }
    }

    /* Sort detailslist by VOMS */
    sd_file_sortbyVO(detailslist, vos);

    for (i = 0; i < detailslist->numServiceDetails; i++) {
        SDService *aserv;
        int ret;

        aserv = details_to_service(detailslist->servicedetails[i], exception);
        if (!aserv) {
            SD_freeException(exception);
            continue;
        }

        ret = add_to_list(list, aserv, exception);
        if (ret) {
            sd_file_freeService(aserv);
            sd_file_freeServiceList(list);
            sd_file_freeServiceDetailsList(detailslist);
            return NULL;
        }
    }

    sd_file_freeServiceDetailsList(detailslist);

    if (!list->numServices) {
        g_free(list);
        set_error(exception, "No services of type %s were found", type);
        return NULL;
    }

    return list;
}

void sd_file_freeServiceDataList(SDServiceDataList *serviceDataList) {
    int i;

    if (!serviceDataList)
        return;

    for (i = 0; i < serviceDataList->numItems; i++) {
        g_free(serviceDataList->items[i].key);
        g_free(serviceDataList->items[i].value);
    }
    g_free(serviceDataList->items);
    g_free(serviceDataList);
}

void sd_file_freeService(SDService *service) {
    if (!service)
        return;

    g_free(service->name);
    g_free(service->type);
    g_free(service->endpoint);
    g_free(service->version);
    g_free(service);
}

void sd_file_freeServiceList(SDServiceList *serviceList) {
    int i;

    if (!serviceList)
        return;

    for (i = 0; i < serviceList->numServices; i++)
        sd_file_freeService(serviceList->services[i]);
    g_free(serviceList->services);
    g_free(serviceList);
}

void sd_file_freeServiceDetailsList(SDServiceDetailsList *servicedetailsList) {
    if (!servicedetailsList)
        return;

    g_free(servicedetailsList->servicedetails);
    g_free(servicedetailsList);
}

void sd_file_freeVOList(SDVOList *vos) {
    int i;

    if (!vos)
        return;

    for (i = 0; i < vos->numNames; i++)
        g_free(vos->names[i]);
    g_free(vos->names);
    g_free(vos);
}

void sd_file_freeServiceDetails(SDServiceDetails *details) {
    if (!details)
        return;

    g_free(details->name);
    g_free(details->type);
    g_free(details->endpoint);
    g_free(details->version);

    g_free(details->site);
    g_free(details->wsdl);
    g_free(details->administration);

    sd_file_freeVOList(details->vos);
    sd_file_freeServiceList(details->associatedServices);
    sd_file_freeServiceDataList(details->data);
    g_free(details);
}

void sd_file_sortbyVO(SDServiceDetailsList *servicedetailsList, const SDVOList *vos) {
    int i, *scores;
    if (!servicedetailsList || !servicedetailsList->numServiceDetails
            || !vos || !vos->numNames)
        return;
    scores = g_new0(int, servicedetailsList->numServiceDetails);
    for (i = 0; i < vos->numNames; i++) {
        int j;
        for (j = 0; j < servicedetailsList->numServiceDetails; j++) {
            SDServiceDetails *details = servicedetailsList->servicedetails[j];
            if (details->vos && details->vos->numNames) {
                int k;
                for (k = 0; k < details->vos->numNames; k++)
                    if (!strcmp(vos->names[i], details->vos->names[k])) {
                        scores[j]++;
                        /* Even if the service has the VO defined *
                         * multiple times, we count it only once. */
                        break;
                    }
            }
        }
    }
    for (i = 0; i < servicedetailsList->numServiceDetails; i++) {
        int j;
        for (j = i + 1; j < servicedetailsList->numServiceDetails; j++) {
            if (scores[i] < scores[j]) {
                int tmp = scores[i];
                SDServiceDetails *tmpSD = servicedetailsList->servicedetails[i];
                scores[i] = scores[j];
                scores[j] = tmp;
                servicedetailsList->servicedetails[i] = servicedetailsList->servicedetails[j];
                servicedetailsList->servicedetails[j] = tmpSD;
            }
        }
    }
    g_free(scores);
}


