/*                                                                -*- mode: c; c-basic-offset: 8;  -*-
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 ============================================================================
  Implementation of Glue2 for gfal
 ============================================================================
 */


#ifndef __GNUC__
#define __attribute__(x)
#endif

#include <stdio.h>
#include "Common.h"
#include "Definitions.h"
#include "ServiceDiscoveryIfce.h"

static char *glue2_base = "o=glue";
extern char *vo_global;
extern char **fqan_global;
extern int fqan_size_global;


/*
return Glue2ServiceID as output, 
by looking for an SRM Endpoint with the hostname matching the
EndpointURL and getting the Service ID from the EndpointServiceForeignKey.
*/
int g2_sd_get_service_id(LDAP *ld, const char *host, char **service_id, char **endpoint_id, char *errbuf, int errbufsz)
{
    LDAPMessage *entry = NULL;
    SDException exception;
    LDAPMessage *reply = NULL;
    struct berval **value = NULL;
    char *query = NULL;
    int sav_errno = 0;
    *service_id = NULL;
    *endpoint_id = NULL;
    int rc = 0;

    if (!host) {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_service_id]: Hostname is not specified", errbuf, errbufsz);
        return (-1);
    }

    query = g_strdup_printf(GLUE2_SERVICE_ID, host);
    
    /*fprintf(stderr, "g2_sd_get_service_id=%s\n", query);*/
   
        rc = connect_search_ldap(glue2_base, query, all_attrs, ld, &exception, &reply, errbuf, errbufsz);
        if (rc != LDAP_SUCCESS) {
            if (reply)
                ldap_msgfree(reply);
            if (query)
            	g_free(query);	       		
          return (-1);
        }


        if ((1 > ldap_count_entries(ld, reply))) {
            if (reply)
                ldap_msgfree(reply);
            if (query)
               g_free(query);	       	       
            return (-1);
        }


        for (entry = ldap_first_entry(ld, reply); entry != NULL && rc == 0; entry
                = ldap_next_entry(ld, entry)) {

        if ( (value = ldap_get_values_len(ld, entry, GLUE2EndpointServiceForeignKey))) {
            if (value && value[0]->bv_val) {
                if ((*service_id = strdup(value[0]->bv_val)) == NULL) {
                    sav_errno = errno ? errno : ENOMEM;
                }
                ber_bvecfree(value);
            }
        }
        if ( (value = ldap_get_values_len(ld, entry, GLUE2EndpointID))) {
            if (value && value[0]->bv_val) {
                if ((*endpoint_id = strdup(value[0]->bv_val)) == NULL) {
                    sav_errno = errno ? errno : ENOMEM;
                }
                ber_bvecfree(value);
            }
        }
		
        if (reply)
		ldap_msgfree(reply);        
	if(*service_id)
		break;

     }

    if (query)
        g_free(query);

    if (!*service_id) {
        if (!sav_errno) {
            sav_errno = EINVAL;
        }
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_service_id]: failed", errbuf, errbufsz);
        errno = sav_errno;
        return (-1);
    }
    return (0);
}


/*  get from the BDII the CE Accesspoint for a given SE */

int g2_sd_get_ce_ap(const char *host, char **ce_ap, char *errbuf, int errbufsz) {
    LDAPMessage *entry = NULL;
    SDException exception;
    LDAPMessage *reply = NULL;
    struct berval **value = NULL;
    char *query = NULL;
    int sav_errno = 0;
    *ce_ap = NULL;
    int rc = 0;
    LDAP *ld = NULL;
    char *service_id = NULL;
    char *endpoint_id = NULL;

    if (!host) {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_ce_ap]: Hostname is not specified", errbuf, errbufsz);
        return (-1);
    }



    if ((ld = get_connection(&exception, errbuf, errbufsz)) == NULL) {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_ce_ap]: Cannot connect to BDII", errbuf, errbufsz);
        if (query)
            g_free(query);
        return (-1);
    }

    g2_sd_get_service_id(ld, host, &service_id, &endpoint_id, errbuf, errbufsz);
    if(!service_id)
    {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_ce_ap]: ServiceID cannot be found", errbuf, errbufsz);
	if(ld)
		close_connection(ld);
	return (-1);      
	
    }        

    query = g_strdup_printf(GLUE2_QUERY_CESE_BIND, service_id);
    fprintf(stderr, "g2_sd_get_ce_ap=%s\n", query);


       rc = connect_search_ldap(glue2_base, query, g2_ce_ap_attrs, ld, &exception, &reply, errbuf, errbufsz);
        if (rc != LDAP_SUCCESS) {
           if (reply)
                ldap_msgfree(reply);
            if (query)
               g_free(query);
	    if (ld)
        	close_connection(ld);
	    if(service_id)
	    	free(service_id);		
            sav_errno = errno;
	    rc = -1;
	    goto out;
            
        }



        if ((1 > ldap_count_entries(ld, reply))) {
            if (reply)
                ldap_msgfree(reply);
            if (query)
               g_free(query);
	    if (ld)
        	close_connection(ld);	       
	    if(service_id)
	    	free(service_id);				
            goto out;
        }


        for (entry = ldap_first_entry(ld, reply); entry != NULL && rc == 0; entry
                = ldap_next_entry(ld, entry)) {


        if ((entry = ldap_first_entry(ld, reply)) && (value
                = ldap_get_values_len(ld, entry, GLUE2_ATTR_CE_AP_ATNM))) {
            if (value && value[0]->bv_val) {
                if ((*ce_ap = strdup(value[0]->bv_val)) == NULL) {
                    sav_errno = errno ? errno : ENOMEM;
                }
                ber_bvecfree(value);
            }
        }

        if (reply)
            ldap_msgfree(reply);
        if(*ce_ap)
	    break;
    }

    if (query)
        g_free(query);
    if (ld)
        close_connection(ld);
    if(service_id)
	    	free(service_id);			
    if(endpoint_id)
	    	free(endpoint_id);					

out:
    if (!*ce_ap) {
        if (!sav_errno) {
            sav_errno = EINVAL;
        }
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_ce_ap]: No GLUE2ToComputingServiceID or GLUE2ToComputingServiceLocalPath found.", errbuf, errbufsz);
        errno = sav_errno;
        return (-1);
    }
    return (0);
}

/* generates part of bdii filter with GLUE2PolicyRule */
char *g2_generate_acbr(const char *glueobject) {

    char tmp[64 + 511];
    char *filter = NULL, *vo = NULL;
    int nb_fqan, filterlen, i;

    if ((vo = vo_global) == NULL)
        return (NULL);

    nb_fqan = fqan_size_global;
    filterlen = nb_fqan * (sizeof (tmp)) + 3 * (64 + 255);

    if ((filter = (char *) calloc(filterlen, sizeof (char))) == NULL)
        return (NULL);

    snprintf(filter, filterlen,
            "(| (GLUE2PolicyRule=VO:%s) (GLUE2PolicyRule=vo:%s) ", vo, vo);

    for (i = 0; i < nb_fqan; ++i) {
        if (fqan_global && fqan_global[i]) {
            snprintf(tmp, sizeof (tmp), " (GLUE2PolicyRule=VOMS:%s)", fqan_global[i]);
            strcat(filter, tmp);
        }
    }

    strcat(filter, ")");
    return (filter);
}

/* get from the BDII the endpoint for the LFC */
int g2_sd_get_lfc_endpoint(char **lfc_endpoint, char *errbuf, int errbufsz) {
    LDAPMessage *entry = NULL;
    char *filter = NULL, *filter_tmp = NULL;
    int rc = 0;
    LDAPMessage *reply = NULL;
    struct berval **value = NULL;
    int sav_errno = 0;
    *lfc_endpoint = NULL;
    SDException exception;
    char *query_policy = NULL;
    char *lfc_fk = NULL;
    LDAP *ld = NULL;

    if ((filter_tmp = g2_generate_acbr("GLUE2PolicyID")) == NULL) {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_lfc_endpoint]: No GLUE2PolicyID found", errbuf, errbufsz);
        return (-1);
    }

    if ((ld = get_connection(&exception, errbuf, errbufsz)) == NULL) {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_lfc_endpoint]: Cannot connect to BDII", errbuf, errbufsz);
        return (-1);
    }

    query_policy = g_strdup_printf(GLUE2_QUERY_LFC_POLICY, filter_tmp);
    free(filter_tmp);

    fprintf(stderr, "g2_sd_get_lfc_endpoint1=%s\n", query_policy);

    while (!*lfc_endpoint && !sav_errno) {

        /*get first the policy rule for a particular vo*/
        rc = connect_search_ldap(glue2_base, query_policy, g2_lfc_endpoint_foreign_key_attrs, ld, &exception, &reply, errbuf, errbufsz);
        if (rc != LDAP_SUCCESS) {
            if (reply) {
                ldap_msgfree(reply);
            }
            sav_errno = errno;
            break;
        }
        if ((entry = ldap_first_entry(ld, reply)) && (value = ldap_get_values_len(ld, entry, GLUE2_ACCESS_POLICY_FK))) {
            if (value && value[0]->bv_val) {
                if ((lfc_fk = strdup(value[0]->bv_val)) == NULL)
                    sav_errno = errno ? errno : ENOMEM;
                ldap_value_free_len(value);
            }
        }

        /*now get the lfc endpoint*/
        rc = asprintf(&filter, GLUE2_QUERY_LFC_ENDPOINT, lfc_fk);
        free(lfc_fk);

        fprintf(stderr, "g2_sd_get_lfc_endpoint2=%s\n", filter);

        if (rc < 0) {
            if (reply) {
                ldap_msgfree(reply);
            }
            if (filter)
                free(filter);
            break;
        }

        rc = connect_search_ldap(glue2_base, filter, g2_lfc_endpoint_attrs, ld, &exception, &reply, errbuf, errbufsz);
        if (rc != LDAP_SUCCESS) {
            if (reply) {
                ldap_msgfree(reply);
            }
            if (filter)
                free(filter);
            sav_errno = errno;
            break;
        }
        if ((entry = ldap_first_entry(ld, reply)) && (value = ldap_get_values_len(ld, entry, GLUE2_ATTR_ENDPOINT))) {
            if (value && value[0]->bv_val) {
                if ((*lfc_endpoint = strdup(value[0]->bv_val)) == NULL)
                    sav_errno = errno ? errno : ENOMEM;
                ldap_value_free_len(value);
            }
        }

        if (reply)
            ldap_msgfree(reply);
        if (filter)
            free(filter);
	    
	if(!*lfc_endpoint)
	  break;
    }

    if (query_policy)
        g_free(query_policy);
    if (ld)
        close_connection(ld);


    if (!*lfc_endpoint) {
        if (!sav_errno) {            
            sav_errno = EINVAL;
        }
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_lfc_endpoint]: No LFC Endpoint found", errbuf, errbufsz);
        errno = sav_errno;
        return (-1);
    }

    return 0;
}

int g2_get_voinfo(LDAP *ld, const char *host, const char *spacetokendesc, char **sa_path, char *errbuf, int errbufsz) {

    int i = 0;
    LDAPMessage *entry = NULL;
    char *filter = NULL, *filter_tmp = NULL;
    int rc = 0;
    LDAPMessage *reply = NULL;
    struct berval **value = NULL;
    int sav_errno = 0;
    SDException exception;
    char *vo = NULL;
    storages *storage_shares = NULL;
    storages *storage_shares_temp = NULL;
    GSList* list = NULL;
    char *GLUE2ShareID = NULL;
    char *GLUE2StorageSharePath = NULL;
    int list_length = 0;
    char *service_id = NULL;
    char *endpoint_id = NULL;
    
    g_setenv("G_SLICE", "always-malloc", 1);

    if ((vo = vo_global) == NULL) {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_get_voinfo]: Invalid Arguments.", errbuf, errbufsz);
        errno = EINVAL;
        return (-1);
    }

    if (!host || !sa_path) {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_get_voinfo]: Invalid Arguments.", errbuf, errbufsz);
        errno = EINVAL;
        return (-1);
    }

    g2_sd_get_service_id(ld, host, &service_id, &endpoint_id, errbuf, errbufsz);
    if(!service_id)
    {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_get_voinfo]: ServiceID cannot be found", errbuf, errbufsz);
	if(ld)
		close_connection(ld);
	return (-1);      
	
    }    


    if (spacetokendesc)
        rc = asprintf(&filter, GLUE2_STORAGE_PATH, service_id, spacetokendesc);
    else
        rc = asprintf(&filter, GLUE2_STORAGE_PATH_NO_ST, service_id);

    fprintf(stderr, "g2_sd_get_storage_path1=%s\n", filter);

    if (rc < 0) return (-1);

    *sa_path = NULL;

    while (!*sa_path && !sav_errno) {

        rc = connect_search_ldap(glue2_base, filter, g2_storage_path_attrs, ld, &exception, &reply, errbuf, errbufsz);
        if (rc != LDAP_SUCCESS) {
            if (reply) {
                ldap_msgfree(reply);
            }
            if (filter)
                free(filter);
            sav_errno = errno;
            goto out;
        }

        if ((1 > ldap_count_entries(ld, reply))) {
            if (reply)
                ldap_msgfree(reply);
            if (filter)
                free(filter);
            goto out;
        }

        for (entry = ldap_first_entry(ld, reply); entry != NULL && rc == 0; entry
                = ldap_next_entry(ld, entry)) {

            if ((value = ldap_get_values_len(ld, entry, GLUE2_GLUE2ShareID)) != NULL) { /*get GLUE2_GLUE2ShareID*/
                if (value && value[0]->bv_val) {
                    GLUE2ShareID = strdup(value[0]->bv_val);
                    if (value)
                        ber_bvecfree(value);
                }
            }


            if ((value = ldap_get_values_len(ld, entry, GLUE2_GLUE2StorageSharePath)) != NULL) { /*get GLUE2_GLUE2ShareID*/
                if (value && value[0]->bv_val) {
                    GLUE2StorageSharePath = strdup(value[0]->bv_val);
                    if (value)
                        ber_bvecfree(value);
                }
            }

            /*append to the list*/
            if (GLUE2StorageSharePath && GLUE2ShareID) {
                storage_shares = g_new0(storages, 1);
                storage_shares->GLUE2ShareID = GLUE2ShareID;
                storage_shares->GLUE2StorageSharePath = GLUE2StorageSharePath;
                list = g_slist_append(list, storage_shares);

            }

        } /*end for*/


        if (reply)
            ldap_msgfree(reply);
        if (GLUE2StorageSharePath && GLUE2ShareID)
            break;

    } /*end while*/

    if (filter)
        free(filter);

    if (list) {
        list_length = g_slist_length(list);
        for (i = 0; i < list_length; i++) {
            storage_shares_temp = (storages *) g_slist_nth(list, i)->data;
            rc = asprintf(&filter, GLUE2_STORAGE_POLICY, endpoint_id, vo);

            fprintf(stderr, "g2_sd_get_storage_path2=%s\n", filter);

            rc = connect_search_ldap(glue2_base, filter, all_attrs, ld, &exception, &reply, errbuf, errbufsz);
            if (rc != LDAP_SUCCESS) {
                if (reply) {
                    ldap_msgfree(reply);
                }
                sav_errno = errno;
                goto out;
            }

            if ((1 > ldap_count_entries(ld, reply))) {
                if (reply)
                    ldap_msgfree(reply);
                if (filter)
                    free(filter);
                goto out;
            } else {
                if (filter)
                    free(filter);
                if (storage_shares_temp->GLUE2StorageSharePath)
                    *sa_path = strdup(storage_shares_temp->GLUE2StorageSharePath);
            }

        }
    }
out:
    if (list) {
        for (i = 0; i < list_length; i++) {
            storage_shares_temp = (storages *) g_slist_nth(list, i)->data;
            g_free(storage_shares_temp->GLUE2StorageSharePath);
            g_free(storage_shares_temp->GLUE2ShareID);
            g_free(storage_shares_temp);
        }
        g_slist_free(list);
    }
    
    if(service_id)
    	free(service_id);
    if(endpoint_id)
    	free(endpoint_id);
		
    if (!*sa_path) {
        if (!sav_errno) {
            if (spacetokendesc) {
                SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_storage_path]: no GLUE2StorageSharePath information found about tag and SE ", errbuf, errbufsz);
            } else {
                SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_storage_path]: no information found about SE  (with no tag) ", errbuf, errbufsz);

            }
            sav_errno = EINVAL;
        }
        errno = sav_errno;
        return (-1);
    }

    return (0);
}

int g2_sd_get_storage_path(const char *host, const char *spacetokendesc,
        char **sa_path, char **sa_root, char *errbuf, int errbufsz) {

    int rc = 0;
    SDException exception;
    LDAP *ld = NULL;

    if (!host || !sa_path) {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_storage_path]: Invalid Arguments passed to retrieve storage path", errbuf, errbufsz);
        errno = EINVAL;
        return (-1);
    }

    if ((ld = get_connection(&exception, errbuf, errbufsz)) == NULL) {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_storage_path]: Cannot connect to BDII", errbuf, errbufsz);
        return (-1);
    }


    *sa_path = *sa_root = NULL;

    rc = g2_get_voinfo(ld, host, spacetokendesc, sa_path, errbuf, errbufsz);

    if (ld)
        close_connection(ld);

    return (rc);
}

/**
 *  get from the BDII the SE type ( disk, srm_v1 )
 *
 *  */

int g2_sd_get_se_types_and_endpoints(const char *host, char ***se_types,
        char ***se_endpoints, char *errbuf, int errbufsz) {
    char host_tmp[255];
    int len_tmp = 0;
    char *port = NULL;
    LDAPMessage *entry = NULL;
    SDException exception;
    char *query = NULL;
    int i, nbentries = 0, n = 0, rc = 0;
    char **sep = NULL, **stp = NULL, **st = NULL, **sv = NULL, **ep = NULL;
    struct berval **value = NULL;
    int sav_errno = 0;
    LDAPMessage *reply = NULL;
    LDAP *ld = NULL;
    char *service_id = NULL;
    char *endpoint_id = NULL;    

    *se_types = NULL;
    *se_endpoints = NULL;

    if (!host) {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_se_types_and_endpoints]: Hostname in empty", errbuf, errbufsz);
        return -1;
    }


    if ((ld = get_connection(&exception, errbuf, errbufsz)) == NULL) {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_se_types_and_endpoints]: Cannot connect to BDII", errbuf, errbufsz);
        return (-1);
    }

    g2_sd_get_service_id(ld, host, &service_id, &endpoint_id, errbuf, errbufsz);
    if(!service_id)
    {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_se_types_and_endpoints]: ServiceID cannot be found", errbuf, errbufsz);
	if(ld)
		close_connection(ld);
	return (-1);      
	
    }       

    strncpy(host_tmp, host, 255);

    if ((port = strchr(host_tmp, ':')) != NULL)
        *port = 0;

    query = g_strdup_printf(GLUE2_QUERY_TYPE_ENDPOINTS, service_id);

    fprintf(stderr, "g2_sd_get_se_types_and_endpoints=%s\n", query);

    while (nbentries < 1 && !sav_errno) {

        rc = connect_search_ldap(glue2_base, query, NULL, ld, &exception, &reply, errbuf, errbufsz);
        if (rc != LDAP_SUCCESS) {
            if (reply) {
                ldap_msgfree(reply);
            }
            rc = -1;
            sav_errno = errno;
            break;
        }

        if ((nbentries = ldap_count_entries(ld, reply)) < 1) {
            SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_se_types_and_endpoints]: No entries found for host.", errbuf, errbufsz);
            if (reply) {
                ldap_msgfree(reply);
            }
            rc = -1;
            break;
        }

        ++nbentries;
        if ((st = calloc(nbentries, sizeof (char *))) == NULL || (sv = calloc(
                nbentries, sizeof (char *))) == NULL || (ep = calloc(nbentries,
                sizeof (char *))) == NULL || (stp = calloc(nbentries,
                sizeof (char *))) == NULL || (sep = calloc(nbentries,
                sizeof (char *))) == NULL) {
            if (reply) {
                ldap_msgfree(reply);
            }

            sav_errno = errno ? errno : ENOMEM;
            rc = -1;
            break;
        }

        for (entry = ldap_first_entry(ld, reply); entry != NULL && rc == 0; entry
                = ldap_next_entry(ld, entry)) {

            if ((value = ldap_get_values_len(ld, entry, GLUE2_ATTR_ENDPOINT)) != NULL) {
                if (value && value[0]->bv_val) {
                    ep[n] = strdup(value[0]->bv_val);
                    if (value)
                        ber_bvecfree(value);

                    if ((value = ldap_get_values_len(ld, entry, GLUE2_ATTR_TYPE)) == NULL) {
                        if (value && ep[n]) {
                            ber_bvecfree(value);
                            free(ep[n]);
                            continue;
                        }
                    }
                    if (st)
                        st[n] = strdup(value[0]->bv_val);
                    if (value)
                        ber_bvecfree(value);

                    if ((value = ldap_get_values_len(ld, entry, GLUE2_ATTR_VERSION))
                            == NULL) {
                        if (value && ep[n] && st[n]) {
                            ber_bvecfree(value);
                            free(ep[n]);
                            free(st[n]);
                            continue;
                        }
                    }
                    if (sv && value && value[0]->bv_val)
                        sv[n] = strdup(value[0]->bv_val);
                    if (value)
                        ber_bvecfree(value);
                    ++n;
                }
            } else {
                if (port == NULL) {
                    /* If port is not yet defined in host_tmp, and is available
                    // it is copied to host_tmp buffer
                    // ... But it will only be used if there is no GLUE2EndpointURL entry */
                    if (value)
                        ber_bvecfree(value);

                    value = ldap_get_values_len(ld, entry, "GLUE2EndpointURL");
                    if (value == NULL) {
                        if (nbentries == 1) {
                            sav_errno = EINVAL;
                            SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_se_types_and_endpoints]: No port published in the BDII for Glue2", errbuf, errbufsz);
                        }
                    } else if (value && value[0]->bv_val && len_tmp + strlen(value[0]->bv_val) + 1 < 255) {
                        port = host_tmp + len_tmp;
                        strcpy(port + 1, value[0]->bv_val);

                    } else {
                        if (value)
                            ber_bvecfree(value);
                        sav_errno = ENAMETOOLONG;
                        if (reply) {
                            ldap_msgfree(reply);
                        }

                        break;
                    }

                }
                if (value)
                    ber_bvecfree(value);
            }
        }
        if (reply) {
            ldap_msgfree(reply);
        }

    }

    if (rc < 0) {
        for (i = 0; i < n; ++i) {
            if (st && st[i])
                free(st[i]);
            if (sv && sv[i])
                free(sv[i]);
            if (ep && ep[i])
                free(ep[i]);
        }
        if (st)
            free(st);
        if (sv)
            free(sv);
        if (ep)
            free(ep);
        if (stp)
            free(stp);
        if (sep)
            free(sep);
        errno = sav_errno ? sav_errno : EINVAL;
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_se_types_and_endpoints]: No available information", errbuf, errbufsz);
    } else {
        if (n > 0) {

            for (i = 0; i < n; ++i) {
                if ((strcmp(st[i], "srm_v1") == 0 || strcmp(st[i], "srm_v2")
                        == 0) && ep[i]) {
                    stp[i] = strdup(st[i]);
                    sep[i] = strdup(ep[i]);
                } else if ((strcasecmp(st[i], "SRM") == 0) && (strncmp(sv[i],
                        "1.", 2)) == 0 && ep[i]) {
                    stp[i] = strdup("srm_v1");
                    sep[i] = strdup(ep[i]);
                } else if ((strcasecmp(st[i], "SRM") == 0) && (strncmp(sv[i],
                        "2.2", 3)) == 0 && ep[i]) {
                    stp[i] = strdup("srm_v2");
                    sep[i] = strdup(ep[i]);
                }
                free(st[i]);
                free(sv[i]);
                free(ep[i]);
            }
            free(st);
            free(sv);
            free(ep);
        } else if (n == 0 && port != NULL && stp && sep) {
            /* There were no GLUE2EndpointID entries...
            // ... so endpoint is hostname:port, and type is disk */
            *port = ':';
            stp[0] = strdup("disk");
            sep[0] = strdup(host_tmp);
        } else {
            free(stp);
            free(sep);
            rc = -1;
            SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_se_types_and_endpoints]: No available information found for SE", errbuf, errbufsz);
            errno = EINVAL;

        }

        if (rc == 0) {
            *se_types = stp;
            *se_endpoints = sep;
        }
    }

    if (ld)
        close_connection(ld);
    if (query)
        g_free(query);
    if (service_id)
    	free(service_id);	 
    if (endpoint_id)
    	free(endpoint_id);	 

    return (rc);
}

/* get from the BDII the list of supported protocols with their associated
 * port number
 */
int g2_sd_get_seap_info(const char *host, char ***access_protocol, int **port, char *errbuf, int errbufsz) {
    char **ap = NULL, *attr = NULL;
    char *query = NULL;
    struct berval **value = NULL;
    BerElement *ber = NULL;
    int nbentries, i = 0, n = 0, rc = 0;
    int *pn = NULL;
    LDAPMessage *entry = NULL, *reply = NULL;
    SDException exception;
    int sav_errno = 0;
    char *tempUrl = NULL;
    char *base_scheme = NULL;
    char *base_host = NULL;
    char *base_path = NULL;
    int base_port = 0;
    int len_tmp = 0;
    LDAP *ld = NULL;
    char *service_id = NULL;
    char *endpoint_id = NULL;    

    if (!host) {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_seap_info]: Hostname in null", errbuf, errbufsz);
        return -1;
    }


    if ((ld = get_connection(&exception, errbuf, errbufsz)) == NULL) {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_seap_info]: Cannot connect to BDII", errbuf, errbufsz);
        return (-1);
    }

    
    g2_sd_get_service_id(ld, host, &service_id, &endpoint_id, errbuf, errbufsz);
    if(!service_id)
    {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_seap_info]: ServiceID cannot be found", errbuf, errbufsz);
	if(ld)
		close_connection(ld);	
	return (-1);      
	
    }       

    /*prepare the query */
    query = g_strdup_printf(GLUE2_QUERY_PROTO_PORT_TYPE, service_id);

    fprintf(stderr, "g2_sd_get_seap_info=%s\n", query);

    while (n < 1 && !sav_errno) {

        /*send query to bdii server*/
        rc = connect_search_ldap(glue2_base, query, g2_seap_attrs, ld, &exception, &reply, errbuf, errbufsz);
        if (rc != LDAP_SUCCESS) {
            if (reply) {
                ldap_msgfree(reply);
            }
            rc = -1;
            sav_errno = errno;
            break;
        }

        /*count entry # */
        if ((nbentries = ldap_count_entries(ld, reply)) < 1) {
            SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_seap_info]: No entries for host.", errbuf, errbufsz);
            if (reply) {
                ldap_msgfree(reply);
            }
            sav_errno = errno ? errno : ENOMEM;
            rc = -1;
            break;
        }


        nbentries++;
        /*allocate memory */
        if ((ap = (char **) calloc(nbentries, sizeof (char *))) == NULL || (pn
                = (int *) calloc(nbentries, sizeof (int))) == NULL) {
            if (reply) {
                ldap_msgfree(reply);
            }

            sav_errno = errno ? errno : ENOMEM;
            rc = -1;
            break;
        }


        for (entry = ldap_first_entry(ld, reply), n = 0; entry != NULL && rc
                == 0; entry = ldap_next_entry(ld, entry), ++n) {

            for (attr = ldap_first_attribute(ld, entry, &ber); attr != NULL
                    && rc == 0; attr = ldap_next_attribute(ld, entry, ber)) {

                if (!ap[n] && (value = ldap_get_values_len(ld, entry, attr))) {
                    if (value && value[0]->bv_val) {
                        if (strcmp(attr, "GLUE2EndpointURL") == 0) {
                            if ((ap[n] = strdup(value[0]->bv_val)) == NULL) {
                                rc = -1;
                                sav_errno = errno ? errno : ENOMEM;
                            } else {
                                tempUrl = ap[n];
                                if (tempUrl)
                                    parse_url(tempUrl, &base_scheme, &base_host, &base_port, &base_path);
                                if (base_scheme && base_port > 0) {
				    free(ap[n]);
                                    ap[n] = base_scheme; /* protocol */
                                    pn[n] = base_port; /* port */
                                } else {
                                    rc = -1;
                                    sav_errno = errno ? errno : ENOMEM;
                                }
                            }
                            if (value)
                                ber_bvecfree(value);
                        }
                    }
                }
                if (attr)
                    ldap_memfree(attr);
            }
            if (ber) {
                ber_free(ber, 0);
                ber = NULL;
            }
        }
        if (reply) {
            ldap_msgfree(reply);
        }

        if (base_host)
            free(base_host);
        if (base_path)
            free(base_path);
    }


    /* if something went wrong, give the memory free */
    if (rc < 0) {
        for (i = 0; i < n; ++i) {
            if (ap[i]) {
                free(ap[i]);
            }
        }
        if (ap && pn) {
            free(ap);
            free(pn);
        }
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_get_seap_info]: No available information", errbuf, errbufsz);
        errno = sav_errno ? sav_errno : EINVAL;
    } else /*return the values via pointer*/ {
        *access_protocol = ap;
        *port = pn;
    }

    if (query)
        g_free(query);
    if (ld)
        close_connection(ld);
    if (service_id)
    	free(service_id);	
    if (endpoint_id)
    	free(endpoint_id);	

    return (rc);
}

int g2_sd_is_online(char *host, char *errbuf, int errbufsz) {
    char host_tmp[255];
    LDAPMessage *reply = NULL;
    SDException exception;
    LDAP *ld = NULL;
    char *query = NULL;
    int rc = 0;
    int is_production;
    char *service_id = NULL;
    char *endpoint_id = NULL;    

    if (!host) {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_is_online]: Host is empty", errbuf, errbufsz);
        return (-1);
    }

    if ((ld = get_connection(&exception, errbuf, errbufsz)) == NULL) {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_is_online]: Cannot connect to BDII", errbuf, errbufsz);
        return (-1);
    }

    g2_sd_get_service_id(ld, host, &service_id, &endpoint_id, errbuf, errbufsz);
    if(!service_id)
    {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_is_online]: ServiceID cannot be found", errbuf, errbufsz);
	if(ld)
		close_connection(ld);
	return (-1);      
        	
    }        

    strncpy(host_tmp, host, 255);

    query = g_strdup_printf(GLUE2_QUERY_SE_STATUS, service_id);

    fprintf(stderr, "g2_sd_is_online=%s\n", query);

    rc = connect_search_ldap(glue2_base, query, NULL, ld, &exception, &reply, errbuf, errbufsz);
    if (rc != LDAP_SUCCESS) {
        SD_setException(&exception, SDStatus_FAILURE, "[BDII][g2_sd_is_online]: No available information", errbuf, errbufsz);
        goto out;
    }

    if ((ldap_count_entries(ld, reply)) < 1)
        is_production = 0;
    else
        is_production = 1;


out:
    if (service_id)
    	free(service_id);
    if (endpoint_id)
    	free(endpoint_id);	
    if (reply)
        ldap_msgfree(reply);
    if (query)
        g_free(query);
    if (ld)
        close_connection(ld);

    if (is_production == 0)
        return (-1);

    return 0;
}

